from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import os
import unittest

import aaf2
from aaf2 import properties
import common




class AAFCopyTests(unittest.TestCase):

    def compare_objects(self, a, b, skip_some=False):
        for key in a.keys():
            if a[key].format_name() == "Property":
                if key in ("Description", "Name") and skip_some:
                    continue
                else:
                    self.assertEqual(a[key].data, b[key].data, "key = {}".format(key))
            else:
                pa = a[key]
                pb = b[key]

                if isinstance(pa, (properties.StrongRefSetProperty,
                                   properties.WeakRefArrayProperty)):
                    for key, item_a in pa.items():
                        self.assertTrue(key in pb)
                        item_b = pb[key]
                        self.compare_objects(item_a, item_b, skip_some)

                elif isinstance(pa, properties.StrongRefVectorProperty):
                    for item_a, item_b in zip(pa, pb):
                        self.compare_objects(item_a, item_b, skip_some)
                elif isinstance(pa, (properties.StrongRefProperty,
                                     properties.WeakRefProperty)):
                    self.compare_objects(pa.value, pb.value, skip_some)

                elif isinstance(pa, properties.StreamProperty):
                    read_size = a.root.cfb.sector_size
                    sa = pa.open('r')
                    sb = pb.open('r')
                    byte_size = sa.dir.byte_size
                    self.assertEqual(byte_size, sb.dir.byte_size)
                    while byte_size > 0:
                        da = sa.read(read_size)
                        db = sb.read(read_size)
                        self.assertEqual(da, db)
                        byte_size -= len(da)

    def test_mob_copy(self):
        test_file = os.path.join(common.test_files_dir(),"test_file_01.aaf")
        result_file = common.get_test_file('copy_all_mobs.aaf')

        with aaf2.open(test_file, 'r') as a:
            with aaf2.open(result_file, 'w', extensions=False) as b:
                b.dictionary.update(a.dictionary)
                for mob in a.content.mobs:
                    b.content.mobs.append(mob.copy(root=b))

        with aaf2.open(test_file, 'r') as a:
            with aaf2.open(result_file, 'r', extensions=False) as b:
                for mob in a.content.mobs:
                    other_mob = a.content.mobs.get(mob.mob_id, None)
                    self.assertEqual(mob.mob_id, other_mob.mob_id)
                    self.compare_objects(mob, other_mob)

                self.compare_objects(a.dictionary, b.dictionary, skip_some=True)

        result_file = common.get_test_file('copy_mobs_fail.aaf')
        with aaf2.open(test_file, 'r') as a:
            with aaf2.open(result_file, 'w', extensions=False) as b:
                with self.assertRaises(aaf2.exceptions.AAFAttachError):
                    b.content.mobs.append(mob.copy())

    def test_copy_toplevel(self):
        test_file = os.path.join(common.test_files_dir(),"test_file_01.aaf")
        result_file = common.get_test_file('copy_toplevel_mobs.aaf')

        mob_ids = []
        with aaf2.open(test_file, 'r') as a:
            with aaf2.open(result_file, 'w', extensions=False) as b:
                b.dictionary.update(a.dictionary)
                mobs = []

                # gather all the dependant mobs
                for mob in a.content.toplevel():
                    mobs.append(mob)
                    mobs.extend(mob.dependant_mobs())

                for mob in mobs:
                    mob_ids.append(mob.mob_id)
                    b.content.mobs.append(mob.copy(root=b))

        with aaf2.open(test_file, 'r') as a:
            with aaf2.open(result_file, 'r', extensions=False) as b:
                for mob_id in mob_ids:
                    self.assertTrue(mob_id in b.content.mobs)
                    mob_a = a.content.mobs[mob_id]
                    mob_b = a.content.mobs[mob_id]
                    self.compare_objects(mob_a, mob_b)

                self.compare_objects(a.dictionary, b.dictionary, skip_some=True)


    def test_copy_slots(self):
        src_file = common.get_test_file('copy_src_slots.aaf')
        dst_file = common.get_test_file('copy_dst_slots.aaf')

        src_mob_id = None
        dst_mob_id = None
        video_rate = "25"

        with aaf2.open(src_file, 'w') as f:

            comp_mob = f.create.CompositionMob()
            comp_mob.name = "src_comp"
            src_mob_id = comp_mob.mob_id
            sequence = f.create.Sequence(media_kind="picture")
            timeline_slot = comp_mob.create_timeline_slot(video_rate)
            timeline_slot.segment = sequence
            for i in range(10):
                sequence.components.append(f.create.Filler('picture', 100))

            sequence = f.create.Sequence(media_kind="sound")
            timeline_slot = comp_mob.create_timeline_slot(video_rate)
            timeline_slot.segment = sequence
            for i in range(10):
                sequence.components.append(f.create.Filler('sound', 100))

            f.content.mobs.append(comp_mob)

        with aaf2.open(src_file, 'r') as a:
            with aaf2.open(dst_file, 'w') as b:
                dst_comp_mob = b.create.CompositionMob()
                dst_comp_mob.name = 'dest_comp'
                dst_mob_id = dst_comp_mob.mob_id
                b.content.mobs.append(dst_comp_mob)

                src_comp_mob = a.content.mobs.get(src_mob_id, None)
                self.assertIsNotNone(src_comp_mob)

                for slot in src_comp_mob.slots:
                    dst_comp_mob.slots.append(slot.copy(root=b))

        with aaf2.open(src_file, 'r') as a:
            with aaf2.open(dst_file, 'r') as b:
                src_comp_mob = a.content.mobs.get(src_mob_id, None)
                self.assertIsNotNone(src_comp_mob)

                dst_comp_mob = b.content.mobs.get(dst_mob_id, None)
                self.assertIsNotNone(dst_comp_mob)

                self.assertNotEqual(src_comp_mob.name, dst_comp_mob.name)

                for slot_a, slot_b in zip(src_comp_mob.slots, dst_comp_mob.slots):
                    self.compare_objects(slot_a, slot_b)



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import os
import aaf2

import unittest

import common
import datetime

class ComponentAAFTests(unittest.TestCase):


    def test_marker(self):
        result_file = common.get_test_file('descriptive_marker.aaf')
        described_slots = set([1,2,3,4])

        with aaf2.open(result_file, 'w') as f:
            ems = f.create.EventMobSlot()
            ems['EditRate'].value = '25'
            ems['SlotID'].value = 1000
            # doesn't work in avid unless you specify
            # the same PhysicalTrackNumber as the target TimelineMobSlot.
            ems['PhysicalTrackNumber'].value = 1

            sequence = f.create.Sequence("DescriptiveMetadata")
            marker = f.create.DescriptiveMarker()
            marker['DescribedSlots'].value = described_slots
            marker['Position'].value = 100
            marker['Comment'].value = "This is a comment"
            marker['CommentMarkerUser'].value = "username"

            sequence.components.append(marker)
            ems.segment = sequence

            mob = f.create.CompositionMob()
            mob.name = "marker_sequence"
            p_slot = mob.create_picture_slot()
            filler  = f.create.Filler()
            filler.media_kind = "picture"
            filler.length = 500
            p_slot.segment.components.append(filler)
            p_slot['PhysicalTrackNumber'].value = 1

            mob.slots.append(ems)
            f.content.mobs.append(mob)

        with aaf2.open(result_file, 'r') as f:

            mob = next(f.content.mobs.values())
            slot = mob.slots.value[1]
            marker = slot.segment.components.value[0]

            assert marker['DescribedSlots'].value == described_slots
            assert marker['Position'].value == 100




if __name__ == "__main__":
    unittest.main()



import os
from setuptools import setup, find_packages
import setuptools.command.build_py

PROJECT_METADATA = {
    "version": "1.7.1",
    "author": 'Mark Reid',
    "author_email": 'mindmark@gmail.com',
    "license": 'MIT',
}

METADATA_TEMPLATE = """
__version__ = "{version}"
__author__ = "{author}"
__author_email__ = "{author_email}"
__license__ = "{license}"
"""


class AddMetadata(setuptools.command.build_py.build_py):
    """Stamps PROJECT_METADATA into __init__ files."""

    def run(self):
        setuptools.command.build_py.build_py.run(self)

        if self.dry_run:
            return

        target_file = os.path.join(self.build_lib, 'aaf2', "__init__.py")
        source_file = os.path.join(os.path.dirname(__file__), 'src', 'aaf2', "__init__.py")

        # get the base data from the original file
        with open(source_file, 'r') as fi:
            src_data = fi.read()

        # write that + the suffix to the target file
        with open(target_file, 'w') as fo:
            fo.write(src_data)
            fo.write(METADATA_TEMPLATE.format(**PROJECT_METADATA))

setup(
    name='pyaaf2',
    description='A python module for reading and writing advanced authoring format files',
    url='https://github.com/markreidvfx/pyaaf2',
    project_urls={
        'Source':
            'https://github.com/markreidvfx/pyaaf2',
        'Documentation':
            'http://pyaaf.readthedocs.io',
        'Issues':
            'https://github.com/markreidvfx/pyaaf2/issues',
    },

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Multimedia :: Video',
        'Topic :: Multimedia :: Video :: Display',
        'Topic :: Multimedia :: Video :: Non-Linear Editor',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
        'Natural Language :: English',
    ],
    keywords='film tv editing editorial edit non-linear edl time',

    platforms='any',

    packages=find_packages(where='src'),
    package_dir={'': 'src'},

    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*',

    cmdclass={'build_py': AddMetadata},

    **PROJECT_METADATA
)



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import os
import aaf2
import common
import unittest

from aaf2.misc import generate_offset_map

def find_retime(f):
    comp = next(f.content.toplevel())
    # comp.dump()
    video_tracks = []
    for component in [item.segment for item in comp.slots]:
        # print(component.media_kind)
        if component.media_kind == 'Picture':
            video_tracks.append(component)

    seq = video_tracks[0]

    spatial_adapter = list(seq.components)[0]
    op_group = list(list(spatial_adapter.segments)[0].components)[0]
    speed_map = None
    offset_map = None

    for param in op_group.parameters:

        if param.name == "PARAM_SPEED_MAP_U":
            speed_map = param
        if param.name == 'PARAM_SPEED_OFFSET_MAP_U':
            offset_map = param

    return (op_group, spatial_adapter), speed_map, offset_map


def compare_speedmap_to_offset_map(path):

    with aaf2.open(path) as f:
        op_group, speed_map, offset_map = find_retime(f)
        start = int(offset_map['PointList'].value[0].time)

        length = op_group[0].length

        error_list = []

        for t, v in generate_offset_map(speed_map, start, length):
            target_value = offset_map.value_at(t)
            error_list.append(abs(target_value - v))

        # print("average error:", sum(error_list) / len(error_list))

        return sum(error_list) / len(error_list)

def error_ok(value, path):
    # print(os.path.basename(path), 'error =', value)
    if value > 1.0e-07:
        return False
    return True

class TestRetime(unittest.TestCase):

    def test_speedmap_step(self):
        test_file = os.path.join(common.test_files_dir(), 'retimes/step01.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))
        test_file = os.path.join(common.test_files_dir(), 'retimes/step02.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))
        test_file = os.path.join(common.test_files_dir(), 'retimes/step03.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))

    def test_speedmap_linear(self):
        test_file = os.path.join(common.test_files_dir(), 'retimes/linear01.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))
        test_file = os.path.join(common.test_files_dir(), 'retimes/linear02.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))
        test_file = os.path.join(common.test_files_dir(), 'retimes/linear03.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))

    def test_speedmap_spline(self):
        test_file = os.path.join(common.test_files_dir(), 'retimes/spline01.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))
        test_file = os.path.join(common.test_files_dir(), 'retimes/spline02.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))
        test_file = os.path.join(common.test_files_dir(), 'retimes/spline03.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))

    def test_speedmap_bezier(self):
        test_file = os.path.join(common.test_files_dir(), 'retimes/bezier01.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))
        test_file = os.path.join(common.test_files_dir(), 'retimes/bezier02.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))
        test_file = os.path.join(common.test_files_dir(), 'retimes/bezier03.aaf')
        error = compare_speedmap_to_offset_map(test_file)
        self.assertTrue(error_ok(error, test_file))

    def skip_test_retime_manually(self):

        test_file = os.path.join(common.test_files_dir(), 'retimes/spline02.aaf')
        test_file = os.path.join(common.test_files_dir(), 'retimes/spline04.aaf')
        # compare_speedmap_to_offset_map(test_file)


        # manual testing here

        import matplotlib.pyplot as plt
        import numpy as np

        with aaf2.open(test_file) as f:
            op_group, speed_map, offset_map = find_retime(f)

            target =[[], []]

            p0 = offset_map['PointList'].value[0]
            # p0.dump()
            sp0 = speed_map['PointList'].value[0]
            sp0.dump()

            # p0 = speed_map['PointList'].value[0]

            # for i in range(0, 200):
            #     target[0].append(i)
            #     target[1].append(offset_map.value_at(i))

            for item in offset_map['PointList'].value:
                target[0].append(item.time)
                target[1].append(item.value)

            start = int(p0.time)
            print(p0.time)
            # op_group[1].dump()

            # print(end, p_last.time)
            pos = 0
            error_list = []

            offset_map_gen = generate_offset_map(speed_map, start, op_group[0].length)
            calcuated = [[], []]
            for t,v in offset_map_gen:
                calcuated[0].append(t)
                calcuated[1].append(v)
                tar = offset_map.value_at(t)
                error = abs(tar - v)
                print(t, error, tar, pos)
                error_list.append(error)

            print("average error:", sum(error_list) / len(error_list))


            plt.plot(target[0], target[1])
            plt.plot(calcuated[0], calcuated[1])

            plt.show()


if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()
    # test_spline2d()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import logging

import sys
import os
import io
import math
import weakref
from array import array
from struct import Struct
import struct
from collections import deque
import random

from .utils import (
    read_u8, read_u16le,
    read_u32le, read_u64le,
    read_filetime, read_sid,
    write_u8, write_u16le,
    write_u32le, write_u64le,
    write_filetime, write_sid,
    decode_utf16le,
    decode_sid, encode_sid,
    unpack_u16le_from, unpack_u32le_from, unpack_u64le_from
)
from .exceptions import CompoundFileBinaryError
from .cache import LRUCacheDict
from .import auid

from io import BytesIO

sentinel = object()

dir_types = {0x00 : 'empty',
             0x01 : 'storage',
             0x02 : 'stream',
             0x03 : 'lock bytes',
             0x04 : 'property',
             0x05 : 'root storage'}

DIFSECT    = 0xFFFFFFFC
FATSECT    = 0xFFFFFFFD
ENDOFCHAIN = 0xFFFFFFFE
FREESECT   = 0xFFFFFFFF

RANGELOCKSECT = (0x7FFFFF00 // 4096) - 1

MAXREGSECT = 0xFFFFFFFA
MAXREGSID  = 0xFFFFFFFA
MAX_DIR_ENTRIES = 0x00FFFFFF

fat_sector_types = {DIFSECT    : "DIFSECT",
                    FATSECT    : "FATSECT",
                    ENDOFCHAIN : "ENDOFCHAIN",
                    FREESECT   : "FREESECT"}

DIR_STRUCT = Struct(str(''.join(( '<',
    '64s', # name 0
    'H',   # name_size 64
    'B',   # dir_type 66
    'B',   # color 67
    'I',   # left_id 68
    'I',   # right_id 72
    'I',   # child_id 76
    '16s', # class_id 80
    'I',   # flags 96
    'Q',   # create_time 100
    'Q',   # modify_time 108
    'I',   # sector_id 116
    'Q',   # byte_size 120
))))

def pretty_sectors(fat):
    return [fat_sector_types.get(item, item) for item in fat]

class Stream(object):
    __slots__ = ('storage', 'dir', 'mode', 'pos', 'fat_chain')
    def __init__(self, storage, entry, mode='r'):
        self.storage = storage
        self.dir = entry
        self.mode = mode
        self.pos = 0
        self.fat_chain = []
        if not mode in ('r', 'w', 'rw'):
            raise ValueError("invalid mode: %s" % mode)
        if self.dir.sector_id is not None:
            self.fat_chain.extend(self.storage.get_fat_chain(self.dir.sector_id, self.is_mini_stream()))

    def tell(self):
        return self.pos

    def seek(self, offset, whence=io.SEEK_SET):
        if whence == io.SEEK_CUR:
            offset = self.tell() + offset
        elif whence == io.SEEK_END:
            offset = self.dir.byte_size + offset
        if offset < 0:
            raise ValueError('New position is before the start of the stream')

        if offset > self.dir.byte_size:
            # logging.debug("overseek %d bytes, padding with zeros" % (offset - self.dir.byte_size))
            self.pos = self.dir.byte_size
            bytes_left = offset - self.dir.byte_size
            min_seek_size = self.storage.sector_size * 4
            while bytes_left:
                bytes_to_write = min(min_seek_size, offset - self.dir.byte_size)
                zeros = bytearray(bytes_to_write)
                self.write(zeros)
                bytes_left -= bytes_to_write

        self.pos = offset
        return offset

    def is_mini_stream(self):
        if self.dir.type == 'root storage':
            return False
        return self.dir.byte_size < self.storage.min_stream_max_size

    def sector_size(self):
        if self.is_mini_stream():
            return self.storage.mini_stream_sector_size
        else:
            return self.storage.sector_size

    def sector_offset(self):
        return self.pos % self.sector_size()

    def sector_index(self):
        return self.pos // self.sector_size()

    def read(self, n=-1):

        byte_size = self.dir.byte_size
        if n == -1:
            bytes_to_read = max(0, byte_size - self.tell())
        else:
            bytes_to_read = max(0, min(n, byte_size - self.tell()))

        result = bytearray(bytes_to_read)
        mv = memoryview(result)

        is_mini_stream = byte_size < self.storage.min_stream_max_size
        full_sector_size = self.storage.sector_size
        mini_sector_size = self.storage.mini_stream_sector_size
        mini_stream_chain = self.storage.mini_stream_chain
        read_sector_data = self.storage.read_sector_data
        sector_data = None
        prev_sid = -1

        if is_mini_stream:
            mini_fat_index     = self.pos // mini_sector_size
            mini_sector_offset = self.pos  % mini_sector_size
            sector_size = mini_sector_size
        else:
            index      = self.pos // full_sector_size
            start_offset = self.pos  % full_sector_size
            sector_size = full_sector_size

        while bytes_to_read > 0:

            # inlined on purpose this loop runs a lot
            if is_mini_stream:
                mini_stream_sid = self.fat_chain[mini_fat_index]
                mini_stream_pos = (mini_stream_sid * mini_sector_size) + mini_sector_offset

                index      = mini_stream_pos // full_sector_size
                sid_offset = mini_stream_pos  % full_sector_size

                sid = mini_stream_chain[index]
                sector_offset = mini_sector_offset

                mini_sector_offset = 0
                mini_fat_index += 1

                if sid != prev_sid:
                    sector_data = read_sector_data(sid)
                    prev_sid = sid

            else:
                sid = self.fat_chain[index]
                sector_offset = start_offset
                sid_offset = start_offset

                index += 1
                start_offset = 0

                sector_data = read_sector_data(sid)

            bytes_can_read = min(bytes_to_read, sector_size - sector_offset)
            assert bytes_can_read > 0

            mv[:bytes_can_read] = sector_data[sid_offset:sid_offset+bytes_can_read]

            self.pos += bytes_can_read
            mv = mv[bytes_can_read:]

            bytes_to_read -= bytes_can_read

        return result

    def allocate(self, byte_size):

        minifat = self.is_mini_stream()
        realloc_data = None
        orig_pos = None
        # convert from minifat to fat
        if minifat and byte_size >= self.storage.min_stream_max_size:
            # logging.debug("converting stream for minifat to fat")
            orig_pos = self.pos
            self.seek(0)
            realloc_data = self.read()
            assert len(realloc_data) == self.dir.byte_size
            self.storage.free_fat_chain(self.dir.sector_id, True)
            self.dir.sector_id = None
            minifat = False
            self.fat_chain = []

        self.dir.byte_size = byte_size
        sector_size = self.sector_size()
        sector_count = (byte_size + sector_size - 1) // sector_size

        current_sects= len(self.fat_chain)
        # logging.debug("%d bytes requires %d sectors at %d has %d" % (byte_size, sector_count, self.sector_size(), current_sects))

        while len(self.fat_chain) < sector_count:
            last_sector_id = self.fat_chain[-1] if self.fat_chain else None
            sid = self.storage.fat_chain_append(last_sector_id, minifat)
            self.fat_chain.append(sid)
            if self.dir.sector_id is None:
                self.dir.sector_id = sid

        if realloc_data is not None:
            self.pos = 0
            self.write(realloc_data)
            self.pos = min(orig_pos, len(realloc_data))

    def write(self, data):
        data_size = len(data)
        current_size = self.dir.byte_size
        new_size = max(self.tell() + data_size, current_size)
        if new_size > current_size:
            self.allocate(new_size)

        mv = memoryview(data)

        is_mini_stream = self.dir.byte_size < self.storage.min_stream_max_size
        full_sector_size = self.storage.sector_size
        mini_sector_size = self.storage.mini_stream_sector_size
        mini_stream_chain = self.storage.mini_stream_chain
        sector_cache = self.storage.sector_cache
        f = self.storage.f

        if is_mini_stream:
            mini_fat_index     = self.pos // mini_sector_size
            mini_sector_offset = self.pos  % mini_sector_size
            sector_size = mini_sector_size
        else:
            index      = self.pos // full_sector_size
            sid_offset = self.pos  % full_sector_size
            sector_size = full_sector_size

        while data_size > 0:

            # inlined on purpose this method can get called a lot
            if is_mini_stream:
                mini_stream_sid = self.fat_chain[mini_fat_index]
                mini_stream_pos  = (mini_stream_sid * mini_sector_size) + mini_sector_offset

                index      = mini_stream_pos // full_sector_size
                sid_offset = mini_stream_pos  % full_sector_size

                sid = mini_stream_chain[index]

                sector_offset = mini_sector_offset
                seek_pos = ((sid + 1) *  full_sector_size) + sid_offset

                mini_fat_index += 1
                mini_sector_offset = 0

            else:
                sid = self.fat_chain[index]
                sector_offset = sid_offset

                seek_pos = ((sid + 1) *  full_sector_size) + sid_offset

                index += 1
                sid_offset = 0

            byte_writeable = min(len(mv), sector_size - sector_offset)
            assert byte_writeable > 0

            if sid in sector_cache:
                del sector_cache[sid]

            f.seek(seek_pos)
            f.write(mv[:byte_writeable])
            self.pos += byte_writeable

            mv = mv[byte_writeable:]
            data_size -= byte_writeable

        assert self.pos <= self.dir.byte_size

    def truncate(self, size=None):
        # print("trunc", self.dir.path())
        if size is None:
            size = self.pos

        current_byte_size = self.dir.byte_size
        is_mini_stream = self.is_mini_stream()
        full_sector_size = self.storage.sector_size
        mini_sector_size = self.storage.mini_stream_sector_size

        # free the stream
        if size == 0:
            self.storage.free_fat_chain(self.dir.sector_id, is_mini_stream)
            self.pos = 0
            self.dir.sector_id = None
            self.dir.byte_size = 0
            self.fat_chain = []
            return


        # grown the stream
        if size > current_byte_size:
            self.allocate(size)
            return

        # shrink to mini stream
        if size < self.storage.min_stream_max_size and not is_mini_stream and self.dir.type != 'root storage':
            orig_pos = self.pos
            self.pos = 0

            realloc_data = self.read(size)
            self.storage.free_fat_chain(self.dir.sector_id, False)

            self.pos = 0
            self.dir.sector_id = None
            self.dir.byte_size = 0
            self.fat_chain = []

            self.write(realloc_data)
            self.pos = min(orig_pos, size)
            assert self.dir.byte_size == size
            return

        if is_mini_stream:
            sector_size = mini_sector_size
            fat_table = self.storage.minifat
        else:
            sector_size = full_sector_size
            fat_table = self.storage.fat

        sector_count = (size + sector_size - 1) // sector_size

        if len(self.fat_chain) > sector_count:
            last_sector_id = self.fat_chain[sector_count-1]
            self.storage.free_fat_chain(self.fat_chain[sector_count], is_mini_stream)
            fat_table[last_sector_id] = ENDOFCHAIN
            self.fat_chain = self.fat_chain[:sector_count]

        self.dir.byte_size = size
        self.pos = min(self.pos, size)

    def close(self):
        pass


def is_red(entry):
    if (entry is not None) and entry.red:
        return True
    return False

def is_not_red(entry):
    return not is_red(entry)

def is_parent_of(parent, entry):
    return parent[0] is entry or parent[1] is entry

def validate_rbtree(root):
    if root is None:
        return 1

    left = root.left()
    right = root.right()

    if is_red(root):
        if is_red(left) or is_red(right):
            print("Red violation {}".format(root.path()))
            # raise CompoundFileBinaryError()

    lh = validate_rbtree(left)
    rh = validate_rbtree(right)

    if left is not None and left >= root:
        raise CompoundFileBinaryError("Binary tree violation" )

    if right is not None and right <= root:
        raise CompoundFileBinaryError("Binary tree violation")

    # Black height mismatch
    # cannot guarantee all aaf Implementations use rbtree
    # if lh != 0 and rh != 0 and lh != rh:
    #     print(lh, rh)
    #     raise CompoundFileBinaryError("Black violation {}".format(root.path()))

    if lh != 0 and rh != 0:
        return lh if is_red(root) else lh + 1
    else:
        return 0

def jsw_single(root, direction):
    other_side = 1 - direction

    new_root = root[other_side]

    root[other_side] = new_root[direction]
    new_root[direction] = root

    root.red = True
    new_root.red = False

    return new_root

def jsw_double(root, direction):
    other_side = 1 - direction
    root[other_side] = jsw_single(root[other_side], other_side)
    return jsw_single(root, direction)

def find_entry_parent(root, entry, max_depth):
    parent = None
    node = root
    count = 0
    while node is not None and count < max_depth:
        if node is entry:
            return parent

        direction = 0 if entry < node else 1
        parent = node
        node = node[direction]
        count += 1

    raise CompoundFileBinaryError("Max Depth Exceeded")

def get_entry_path(root, entry, max_depth):
    parent = None
    node = root
    count = 0
    path = []
    while node is not None and count < max_depth:

        path.append(node)
        if node is entry:
            break
        direction = 0 if entry < node else 1
        parent = node
        node = node[direction]
        count += 1

    return path


class DirEntry(object):
    __slots__ = ('storage', 'dir_id', 'parent', 'data', '_name', '__weakref__')

    def __init__(self, storage, dir_id, data=None):
        self.storage = storage

        self.parent = None
        if data is None:
            self.data = bytearray(128)
            # setting dir_id to None disable mark_modified
            self.dir_id = None
            self.left_id = None
            self.right_id = None
            self.child_id = None
            self.sector_id = None
        else:
            self.data = data

        # mark modified will now work
        self.dir_id = dir_id
        self._name = sentinel

    @property
    def name(self):
        if self._name is not sentinel:
            return self._name
        name_size = unpack_u16le_from(self.data, 64)
        assert name_size <= 64
        name =  decode_utf16le(self.data[:name_size])
        self._name = name
        return name

    @name.setter
    def name(self, value):
        name_data = value.encode("utf-16le")
        name_size = len(name_data)
        assert name_size <= 64
        self._name = value
        self.data[:name_size] = name_data
        pad = 64 - name_size
        for i in range(pad):
            self.data[name_size +i] = 0

        # includes null terminator? should re-verify this
        struct.pack_into(str('<H'), self.data, 64, min(name_size+2, 64))
        self.mark_modified()

    @property
    def type(self):
        return dir_types.get(self.data[66] , "unknown")

    @type.setter
    def type(self, value):
        t = None
        for k,v in dir_types.items():
            if v == value:
                t = k
                break
        if t is None:
            raise ValueError("invalid dir type: %s" % str(value))

        self.data[66] = t
        self.mark_modified()

    @property
    def color(self):
        if self.data[67] == 0x01:
            return 'black'
        return 'red'

    @color.setter
    def color(self, value):
        if value == 'black':
            self.data[67] = 0x01
        elif value == 'red':
            self.data[67] = 0x00
        else:
            raise ValueError("invalid dir type: %s" % str(value))

        self.mark_modified()

    @property
    def red(self):
        if self.data[67] == 0x01:
            return False
        return True

    @red.setter
    def red(self, value):
        if value:
            self.data[67] = 0x00
        else:
            self.data[67] = 0x01

        self.mark_modified()

    @property
    def left_id(self):
        sid = unpack_u32le_from(self.data, 68)
        return decode_sid(sid)

    @left_id.setter
    def left_id(self, value):
        struct.pack_into(str('<I'), self.data, 68, encode_sid(value))
        self.mark_modified()

    @property
    def right_id(self):
        sid = unpack_u32le_from(self.data, 72)
        # sid = struct.unpack_from(str('<I'), bytes(self.data), 72)[0]
        return decode_sid(sid)

    @right_id.setter
    def right_id(self, value):
        struct.pack_into(str('<I'), self.data, 72, encode_sid(value))
        self.mark_modified()

    @property
    def child_id(self):
        sid = unpack_u32le_from(self.data, 76)
        return decode_sid(sid)

    @child_id.setter
    def child_id(self, value):
        struct.pack_into(str('<I'), self.data, 76, encode_sid(value))
        self.mark_modified()

    @property
    def class_id(self):
        value = auid.AUID(bytes_le=self.data[80:96])
        if value.int == 0:
            return None
        return value

    @class_id.setter
    def class_id(self, value):
        if value is None:
            self.data[80:96] = bytearray(16)
        else:
            self.data[80:96] = value.bytes_le
        self.mark_modified()

    @property
    def flags(self):
        flags = unpack_u32le_from(self.data, 96)
        return flags

    @flags.setter
    def flags(self, value):
        struct.pack_into(str('<I'), self.data, 96, value)
        self.mark_modified()

    @property
    def create_time(self):
        value = unpack_u64le_from(self.data, 100)
        return value

    @create_time.setter
    def create_time(self, value):
        struct.pack_into(str('<Q'), self.data, 100, value)
        self.mark_modified()

    @property
    def modify_time(self):
        value = unpack_u64le_from(self.data, 108)
        return value

    @modify_time.setter
    def modify_time(self, value):
        struct.pack_into(str('<Q'), bytes(self.data), 108, value)
        self.mark_modified()

    @property
    def sector_id(self):
        sid = unpack_u32le_from(self.data, 116)
        return decode_sid(sid)

    @sector_id.setter
    def sector_id(self, value):
        struct.pack_into(str('<I'), self.data, 116, encode_sid(value))
        self.mark_modified()

    @property
    def byte_size(self):
        value = unpack_u64le_from(self.data, 120)
        return value

    @byte_size.setter
    def byte_size(self, value):
        struct.pack_into(str('<Q'), self.data, 120, value)
        self.mark_modified()

    def mark_modified(self):
        if self.storage.mode in ('r', 'rb'):
            return

        if self.dir_id is None:
            return

        self.storage.modified[self.dir_id] = self
        if len(self.storage.modified) > 128:
            self.storage.write_modified_dir_entries()

    def __lt__(self, other):
        if isinstance(other, DirEntry):
            other = other.name

        if len(self.name) == len(other):
            # compare not case senstive
            return self.name.upper() < other.upper()
        else:
            # shorter names are always less then
            return len(self.name) < len(other)

    def __le__(self, other):
        if self == other:
            return True
        return self < other

    def __gt__(self, other):
        return other < self

    def __ge__(self, other):
        if self == other:
            return True
        return self > other

    def __eq__(self, other):
        if other is None:
            return False

        if isinstance(other, DirEntry):
            other = other.name
        if len(self.name) == len(other):
            return self.name.upper() == other.upper()
        return False

    def __getitem__(self, index):
        return self.left() if index == 0 else self.right()

    def __setitem__(self, index, value):
        if value is None:
            dir_id = None
        else:
            dir_id = value.dir_id

        if index == 0:
            self.left_id = dir_id
        else:
            self.right_id = dir_id

    def left(self):
        return self.storage.read_dir_entry(self.left_id, self.parent)

    def right(self):
        return self.storage.read_dir_entry(self.right_id, self.parent)

    def child(self):
        return self.storage.read_dir_entry(self.child_id, self)

    def add_child(self, entry):
        entry.parent = self
        entry.color = 'black'
        child = self.child()
        if child is None:
            self.child_id = entry.dir_id
        else:
            # child.insert_old(entry)

            # make sure entry is part of cache
            # or insert might now work correctly
            self.storage.dir_cache[entry.dir_id] = entry
            self.insert(entry)

        if self.dir_id in self.storage.children_cache:
            self.storage.children_cache[self.dir_id][entry.name] = entry

    def insert(self, entry):
        """
        Inserts entry into child folder tree.
        Tries to maintain a balanced red black tree.
        Technique is base on topdown insert approach in described in
        https://eternallyconfuzzled.com/red-black-trees-c-the-most-common-balanced-binary-search-tree
        """

        dir_per_sector = self.storage.sector_size // 128
        max_dirs_entries = self.storage.dir_sector_count * dir_per_sector

        head = DirEntry(self.storage, None) # False tree root
        head.red = True
        entry.red = True

        grand_grand_grand_parent = None
        grand_grand_parent = head
        grand_parent = None
        parent = None
        direction = 0
        last = 0

        node = self.child()
        self.child_id = None
        grand_grand_parent.right_id = node.dir_id

        assert node
        count = 0

        while count < max_dirs_entries:

            if node is None:
                node = entry
                parent[direction] = node

            elif is_red(node[0]) and is_red(node[1]):
                # Color flip
                node.red = True
                node[0].red = False
                node[1].red = False

            # Fix red violations
            if is_red(node) and is_red(parent):
                if grand_grand_parent[0] is grand_parent:
                    direction2 = 0
                elif grand_grand_parent[1] is grand_parent:
                    direction2 = 1
                else:
                    raise CompoundFileBinaryError()

                if node is parent[last]:
                    grand_grand_parent[direction2] = jsw_single(grand_parent, 1 - last)
                    # restore parent references
                    # NOTE: Example implementation doesn't do this
                    grand_parent = grand_grand_parent
                    grand_grand_parent = grand_grand_grand_parent

                    # assert is_parent_of(parent, node)
                    # assert is_parent_of(grand_parent, parent)

                elif node is parent[1-last]:
                    grand_grand_parent[direction2] = jsw_double(grand_parent, 1 - last)
                    # restore parent references
                    # NOTE: Example implementation doesn't do this
                    parent = grand_grand_parent
                    if parent is head:
                        grand_parent = None
                    else:
                        grand_parent = grand_grand_grand_parent
                    grand_grand_parent = None

                    # assert is_parent_of(parent, node)
                    # if grand_parent is not None:
                    #     assert is_parent_of(grand_parent, parent)
                else:
                    # can this happen?
                    raise CompoundFileBinaryError()

            # node has been inserted
            if node is entry:
                break

            last = direction
            direction = 0 if entry < node else 1

            if grand_grand_parent is not None:
                grand_grand_grand_parent = grand_grand_parent

            if grand_parent is not None:
                grand_grand_parent = grand_parent

            grand_parent = parent
            parent = node
            node = node[direction]

            assert is_parent_of(parent, node)
            if grand_parent:
                assert is_parent_of(grand_parent, parent)
            if grand_grand_parent:
                assert is_parent_of(grand_grand_parent, grand_parent)

            count += 1

        if count >= max_dirs_entries:
            raise CompoundFileBinaryError("max dir entries limit reached")

        # update root of tree as it could have changed
        self.child_id = head.right_id
        self.child().red = False

    def pop(self):
        """
        remove self from self.parent folder binary search tree.
        Tries to maintain a balanced red black tree.
        Technique is base on topdown remove approach in described in
        https://eternallyconfuzzled.com/red-black-trees-c-the-most-common-balanced-binary-search-tree
        """
        entry = self

        dir_per_sector = self.storage.sector_size // 128
        max_dirs_entries = self.storage.dir_sector_count * dir_per_sector
        count = 0

        head = DirEntry(self.storage, None) # False tree root
        head.red = True
        head.name = "" # NOTE: any name will be less then this
        node = head
        node[1] = self.parent.child()
        grand_parent = None
        parent = None
        entry_parent = None
        entry_grand_parent = None
        direction = 1
        found = None

        # This keeps going until predecessor is found, even if entry is found
        while node[direction] is not None and count < max_dirs_entries:

            last = direction
            grand_parent = parent
            parent = node
            node = node[direction]

            # The trick here is after entry is found
            # this will continue to be used to find the predecessor of entry.
            # its quite clever!
            direction = int(node < entry)

            if node is entry:
                # store the grand parent because the parent of entry
                # can change during rebalance below
                if grand_parent is None:
                    entry_grand_parent = head
                else:
                    entry_grand_parent = grand_parent
                found = node

            # Push the red node down
            if is_not_red(node) and is_not_red(node[direction]):

                if is_red(node[1 - direction]):
                    parent[last] = jsw_single(node, direction)
                    parent = parent[last]
                    assert is_parent_of(parent, node)

                elif is_not_red(node[1 - direction]):
                    sibling = parent[1 - direction]
                    if sibling is not None:

                        if is_not_red(sibling[1 - last]) and is_not_red(sibling[last]):
                            # Color flip
                            parent.red = False
                            sibling.red = True
                            node.red = True
                        else:
                            if grand_parent[0] == parent:
                                direction2 = 0
                            elif  grand_parent[1] == parent:
                                direction2 = 1
                            else:
                                # can this happen?
                                raise CompoundFileBinaryError()

                            if is_red(sibling[last]):
                                grand_parent[direction2] = jsw_double(parent, last)

                            elif is_red(sibling[1 - last]):
                                grand_parent[direction2] = jsw_single(parent, last)

                            # Ensure correct coloring
                            node.red = True
                            grand_parent[direction2].red = True;
                            grand_parent[direction2][0].red = False;
                            grand_parent[direction2][1].red = False;

                            assert is_parent_of(parent, node)

            count += 1

        assert found
        if count >= max_dirs_entries:
            raise CompoundFileBinaryError("max dir entries limit reached")

        # entry parent could have changed during rebalance
        max_search_depth = 4 # grand_parent -> parent -> node
        entry_parent = find_entry_parent(entry_grand_parent, entry, max_search_depth)

        if entry_parent[0] is entry:
            entry_direction = 0
        elif entry_parent[1] is entry:
            entry_direction = 1
        else:
            raise CompoundFileBinaryError("Unable to find entry parent")

        # node is the predecessor. node will be removed and entry will be replaced
        if node is not entry:
            # remove the predecessor
            parent[parent[1] is node] = node[node[0] is None]

            # replace entry with its predecessor
            node[0] = entry[0]
            node[1] = entry[1]
            node.red = entry.red

            entry_parent[entry_direction] = node
        else:
            entry_side = int(entry[0] is None)
            entry_parent[entry_direction] = entry[entry_side]

        # update root of tree as it could have changed
        self.parent.child_id = head.right_id
        if self.parent.child_id:
            self.parent.child().red = False

        # clear from cache
        if self.parent.dir_id in self.storage.children_cache:
            del self.storage.children_cache[self.parent.dir_id][entry.name]

        # clear parent and left and right
        self.left_id = None
        self.right_id = None
        self.parent = None

    def rebalance_children_tree(self):

        children  = self.listdir()
        self.child_id = None
        random.shuffle(children)
        for c in children:
            c.left_id = None
            c.right_id = None
            self.add_child(c)


    def path(self):
        path = []
        parent = self
        while parent:
            name = parent.name
            if name == "Root Entry":
                break
            path.append(parent.name)
            parent= parent.parent
        return  '/' + '/'.join(reversed(path))

    def open(self, mode='r'):
        if self.type != 'stream':
            raise TypeError("can only open streams")
        return self.storage.open(self, mode)

    def isdir(self):
        return self.type in ('storage', 'root storage')

    def isroot(self):
        return self.type == 'root storage'

    def listdir(self):
        return self.storage.listdir(self)

    def makedir(self, relative_path, class_id = None):
        if not self.isdir():
            raise TypeError("can only add a DirEntry to a storage type")
        sep = '/'
        if self.isroot():
            sep = ''

        path = self.path() + sep + relative_path
        return self.storage.makedir(path, class_id)

    def isfile(self):
        return self.type == 'stream'

    def get(self, name, default=None):
        dir_dict = self.storage.listdir_dict(self)
        return dir_dict.get(name, default)

    def touch(self, name):
        item = self.get(name, None)
        if item:
            return item

        sep = '/'
        if self.isroot():
            sep = ''

        path = self.path() + sep + name
        return self.storage.create_dir_entry(path, 'stream', None)

    def write(self):
        f = self.storage.f
        f.seek(self.storage.dir_entry_pos(self.dir_id))
        f.write(self.data)

    def read(self):
        f = self.storage.f
        f.seek(self.storage.dir_entry_pos(self.dir_id))
        f.readinto(self.data)

    def __repr__(self):
        return self.name

def extend_sid_table(f, table, byte_size):
    n = byte_size // 4
    if isinstance(f, io.RawIOBase):
        table.fromfile(f, n)
    elif hasattr(table, 'frombytes'):
        table.frombytes(f.read(byte_size))
    else:
        # try deprecated from string
        table.fromstring(f.read(byte_size))

class CompoundFileBinary(object):
    def __init__(self, file_object, mode='rb', sector_size=4096):

        self.f = file_object

        self.difat = [[]]
        self.fat = array(str('I'))
        self.fat_freelist = []

        self.minifat = array(str('I'))
        self.minifat_freelist = []

        self.difat_chain = []
        self.minifat_chain = []
        self.dir_fat_chain = []

        self.mini_stream_chain = []

        self.modified = {}

        self.sector_cache = LRUCacheDict()
        self.dir_cache = weakref.WeakValueDictionary()
        self.children_cache = LRUCacheDict()
        self.dir_freelist = []

        self.debug_grow = False
        self.is_open = True

        if isinstance(self.f, BytesIO):
            self.mode = 'wb+'
        else:
            self.mode = mode

        if self.mode in ("r", "r+", "rb", 'rb+'):

            self.read_header()
            self.read_fat()
            mini_stream_byte_size = self.read_minifat()

            # create dir_fat_chain and read root dir entry
            self.dir_fat_chain = self.get_fat_chain(self.dir_sector_start)
            if len(self.dir_fat_chain) != self.dir_sector_count:
                logging.info("read dir_sector_count missmatch, using fat chain length")
                self.dir_sector_count = len(self.dir_fat_chain)

            logging.debug("read %d dir sectors" % len(self.dir_fat_chain))
            self.root = self.read_dir_entry(0)
            self.dir_cache[0] = self.root

            # create mini stream fat chain
            if self.minifat_sector_count:
                self.mini_stream_chain = self.get_fat_chain(self.root.sector_id)

            if self.root.sector_id is not None and mini_stream_byte_size != self.root.byte_size:
                message = "mini stream size missmatch: %d != %d, using size from minifat"
                logging.warning(message % (self.root.byte_size, mini_stream_byte_size))
        else:
            self.setup_empty(sector_size)
            self.write_header()

            logging.debug("pos: %d" % self.f.tell())

            logging.debug("writing root dir sector")
            self.root.write()
            self.f.write(bytearray(self.sector_size - 128))
            self.write_fat()

    def close(self):
        if self.mode in ("r", "rb"):
            return

        # calculate mini stream size
        if self.root.sector_id is not None:
            # I cannot find this documented anywhere but the size of the mini stream
            # is the size up to the last mini sector is uses. Not the total Non FREESECT's.
            # If self.root.byte_size is not set correctly the some applications will crash hard...

            # find last non-free sect
            for i,v in enumerate(reversed(self.minifat)):
                if v != FREESECT:
                    break

            last_used_sector_id = len(self.minifat) - i
            mini_stream_byte_size = (last_used_sector_id * self.mini_stream_sector_size)
            self.root.byte_size = mini_stream_byte_size

            # Truncate ministream
            s = Stream(self, self.root, 'rw')
            s.truncate(mini_stream_byte_size)

        self.write_header()
        self.write_difat()
        self.write_fat()
        self.write_minifat()
        self.write_dir_entries()

        # Truncate file to the last free sector
        for i,v in enumerate(reversed(self.fat)):
            if v != FREESECT:
                break

        last_used_sector_id = len(self.fat) - i
        pos = (last_used_sector_id + 1) *  self.sector_size
        self.f.seek(pos)
        self.f.truncate()

        self.is_open = False

    def setup_empty(self, sector_size):

        if sector_size == 4096:
            self.class_id = auid.AUID("0d010201-0200-0000-060e-2b3403020101")
        elif sector_size == 512:
            self.class_id = auid.AUID("42464141-000d-4d4f-060e-2b34010101ff")
        else:
            raise ValueError("sector size must be 4096 or 512")

        self.major_version = 4
        self.minor_version =  62

        self.byte_order = "le"

        self.sector_size = sector_size
        self.mini_stream_sector_size = 64

        self.dir_sector_count = 1
        self.fat_sector_count = 1
        self.dir_sector_start = 0

        self.transaction_signature = 1
        self.min_stream_max_size = 4096

        self.minifat_sector_start = FREESECT
        self.minifat_sector_count = 0

        self.difat_sector_start = FREESECT
        self.difat_sector_count = 0

        self.difat = [[]]
        for i in range(109):
            self.difat[0].append(FREESECT)

        self.difat[0][0] = 1

        for i in range(self.sector_size // 4):
            self.fat.append(FREESECT)
            if i > 1:
                self.fat_freelist.append(i)

        self.fat[0] = ENDOFCHAIN # end of dir chain
        self.fat[self.difat[0][0]] = FATSECT

        self.root = DirEntry(self, 0)
        self.root.name = 'Root Entry'
        self.root.sector_id = None
        self.root.type = 'root storage'
        self.root.class_id = auid.AUID("b3b398a5-1c90-11d4-8053-080036210804")

        self.dir_cache[0] = self.root

        self.dir_fat_chain = [0]

        # raise NotImplementedError("mode: %s supported not implemented" % self.f.mode)

    def write_header(self):
        logging.debug("writiing header")
        f = self.f
        f.seek(0)
        f.write(b'\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1') # Magic
        f.write(self.class_id.bytes_le)
        write_u16le(f, self.minor_version)
        write_u16le(f, self.major_version)
        write_u16le(f, 0xFFFE) # byte order le
        write_u16le(f, int(math.log(self.sector_size, 2)))
        write_u16le(f, int(math.log(self.mini_stream_sector_size, 2)))
        f.write(b'\0' * 6) # skip reserved

        write_u32le(f, self.dir_sector_count)
        write_u32le(f, self.fat_sector_count)
        write_u32le(f, self.dir_sector_start)
        write_u32le(f, self.transaction_signature)
        write_u32le(f, self.min_stream_max_size)

        write_u32le(f, self.minifat_sector_start)
        write_u32le(f, self.minifat_sector_count)

        write_u32le(f, self.difat_sector_start)
        write_u32le(f, self.difat_sector_count)

        for i in range(109):
            write_u32le(f, self.difat[0][i])

        for i in range(self.sector_size - f.tell()):
            f.write(b'\0')


    def read_header(self):

        f = self.f
        f.seek(0)

        self.magic = f.read(8)
        logging.debug("magic: %s" % str([self.magic]))

        if self.magic != b'\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1':
            raise CompoundFileBinaryError("invalid file magic signature: {}".format([self.magic]))

        self.class_id = auid.AUID(bytes_le=f.read(16))
        logging.debug("clsid: %s" % str(self.class_id))

        self.minor_version = read_u16le(f)
        logging.debug("minor_version: %d" % self.minor_version)

        self.major_version = read_u16le(f)
        logging.debug("major_version: %d" % self.major_version)

        byte_order = read_u16le(f)
        if byte_order == 0xFFFE:
            self.byte_order = 'le'
        else:
            raise NotImplementedError("endian format:0x%X not supported" % byte_order)

        logging.debug("byte_order: %s" % self.byte_order)

        size = read_u16le(f)
        self.sector_size = pow(2, size)
        logging.debug("sector_size: %d -> %d" % (size, self.sector_size))

        size = read_u16le(f)
        self.mini_stream_sector_size = pow(2, size)
        logging.debug("mini_stream_sector_size: %d -> %d" % (size, self.mini_stream_sector_size))

        if not self.sector_size in (4096, 512):
            raise NotImplementedError("unsupported sector size: %d" % self.sector_size)
        if self.mini_stream_sector_size != 64:
            raise NotImplementedError("unsupported mini sector size: %d" % self.mini_stream_sector_size)

        f.read(6) # skip reserved

        self.dir_sector_count = read_u32le(f)
        logging.debug("dir_sector_count: %d" % self.dir_sector_count)

        self.fat_sector_count = read_u32le(f)
        logging.debug("fat_sector_count: %d" % self.fat_sector_count)

        self.dir_sector_start = read_u32le(f)
        logging.debug("dir_sector_start: %d" % self.dir_sector_start)

        self.transaction_signature = read_u32le(f)
        logging.debug("transaction_signature: %d" % self.transaction_signature)

        self.min_stream_max_size = read_u32le(f)
        logging.debug("min_stream_max_size: %d" % self.min_stream_max_size)

        self.minifat_sector_start = read_u32le(f)
        logging.debug("minifat_sector_start: %d" % self.minifat_sector_start)

        self.minifat_sector_count = read_u32le(f)
        logging.debug("minifat_sector_count: %d" % self.minifat_sector_count)

        self.difat_sector_start = read_u32le(f)
        logging.debug("difat_sector_start: %d" % self.difat_sector_start)

        self.difat_sector_count = read_u32le(f)
        logging.debug("difat_sector_count: %d" % self.difat_sector_count)

        self.difat = [[]]

        logging.debug("reading header difat at %d" % f.tell())
        for i in range(109):
            item = read_u32le(f)
            self.difat[0].append(item)

        sectors_left = self.difat_sector_count

        sid = self.difat_sector_start

        # reading difat sectors
        while sectors_left:
            logging.debug("reading difat sid: %d", sid)
            sector_type = fat_sector_types.get(sid, sid)
            if not isinstance(sector_type, int):
                break

            self.difat_chain.append(sid)
            f.seek((sid + 1) *  self.sector_size)
            difat = []
            for i in range( (self.sector_size // 4)):
                item = read_u32le(f)
                difat.append(item)
            self.difat.append(difat)

            sid = difat[-1]
            logging.debug("next difat: %d" % sid)
            sectors_left -= 1

    def iter_difat(self):
        for i, sid in enumerate(self.difat[0]):
            yield 0, i, sid

        t = 1
        for item in self.difat[1:]:
            for i, sid in enumerate(item[:-1]):
                yield t, i, sid
            t+=1


    def write_difat(self):
        f = self.f
        # write header entries
        f.seek(76)

        logging.debug("writing header difat")
        for i in range(109):
            write_u32le(f, self.difat[0][i])

        for i in range(self.sector_size - f.tell()):
            f.write(b'\0')

        if self.difat_sector_count == 0:
            return

        sid = self.difat_sector_start
        assert len(self.difat[1:]) == self.difat_sector_count
        for table in self.difat[1:]:

            sector_type = fat_sector_types.get(sid, sid)
            if not isinstance(sector_type, int):
                raise IOError("bad difat sector type")

            pos = (sid + 1) * self.sector_size
            logging.debug("writing difat to sid: %d at: %d" % (sid,pos))
            f.seek(pos)
            for i in range(self.sector_size // 4):
                write_u32le(f, table[i])

            sid = table[-1]

    def read_fat(self):
        f = self.f
        self.fat = array(str('I'))
        sector_count = 0
        fat_sectors = []
        for t, i, sid in self.iter_difat():

            sector_type = fat_sector_types.get(sid, sid)
            if not isinstance(sector_type, int):
                continue

            fat_sectors.append(sid)

        #  len(fat_sectors),self.fat_sector_count
        # assert len(fat_sectors) == self.fat_sector_count
        if len(fat_sectors) != self.fat_sector_count:
            logging.warning("fat sector count missmatch difat: %d header: %d" % (len(fat_sectors), self.fat_sector_count))
            self.fat_sector_count = len(fat_sectors)

        for sid in fat_sectors:
            pos = (sid + 1) *  self.sector_size
            f.seek(pos)
            extend_sid_table(f, self.fat, self.sector_size)
            sector_count += 1

        if sys.byteorder == 'big':
            self.fat.byteswap()

        for i,v in enumerate(self.fat):
            if v == FREESECT:
                self.fat_freelist.append(i)

        logging.debug("read %d fat sectors ", sector_count)

        if self.sector_size == 4096 and len(self.fat) > RANGELOCKSECT:
            if self.fat[RANGELOCKSECT] != ENDOFCHAIN:
                logging.warning("range lock sector has data")

    def write_fat(self):
        logging.debug("writing fat")
        f = self.f
        sector_count = 0

        assert len(self.fat)*4 % self.sector_size == 0

        fat_sectors = []

        for t, i, sid in self.iter_difat():
            sector_type = fat_sector_types.get(sid, sid)
            if not isinstance(sector_type, int):
                continue
            fat_sectors.append(sid)

        # check that the difat has enough entries to hold the current fat
        assert len(fat_sectors) == len(self.fat)*4 // self.sector_size

        element_count = self.sector_size // 4
        fat_table_struct = Struct(str('<%dI' % element_count))
        for i, sid in enumerate(fat_sectors):

            # logging.debug("writing fat to sid: %d" % sid)
            f.seek((sid + 1) *  self.sector_size)
            start = i * element_count
            end = start + element_count
            f.write(fat_table_struct.pack(*self.fat[start:end]))

    def read_minifat(self):
        f = self.f
        sector_count = 0
        self.minifat = array(str('I'))

        for sid in self.get_fat_chain(self.minifat_sector_start):
            self.minifat_chain.append(sid)
            f.seek((sid + 1) *  self.sector_size)
            extend_sid_table(f, self.minifat, self.sector_size)
            sector_count += 1

        if sys.byteorder == 'big':
             self.minifat.byteswap()

        last_used_sector = 0
        for i,v in enumerate(self.minifat):
            if v == FREESECT:
                self.minifat_freelist.append(i)
            else:
                last_used_sector = i

        mini_stream_byte_size = ((last_used_sector+1) * self.mini_stream_sector_size)

        logging.debug("read %d mini fat sectors", sector_count)
        return mini_stream_byte_size

    def write_minifat(self):
        f = self.f
        sector_count = 0

        element_count = self.sector_size // 4
        fat_table_struct = Struct(str('<%dI' % element_count))

        for i, sid in enumerate(self.get_fat_chain(self.minifat_sector_start)):
            pos = (sid + 1) *  self.sector_size
            f.seek(pos)
            start = i * element_count
            end = start + element_count
            f.write(fat_table_struct.pack(*self.minifat[start:end]))

    def write_modified_dir_entries(self):

        f = self.f
        for dir_id in sorted(self.modified):
            entry = self.modified[dir_id]
            stream_pos = entry.dir_id * 128
            chain_index = stream_pos // self.sector_size
            sid_offset  = stream_pos % self.sector_size
            sid = self.dir_fat_chain[chain_index]

            pos = ((sid + 1) *  self.sector_size) + sid_offset

            f.seek(pos)
            # force black everything
            # entry.data[67] = 0x01
            f.write(entry.data)

            # invalidate  sector
            if sid in self.sector_cache:
                del self.sector_cache[sid]

        self.modified = {}

    def write_dir_entries(self):
        self.write_modified_dir_entries()

        # clear empty DirEntrys
        empty_dir = bytearray(128)
        f = self.f

        self.dir_freelist.sort()
        for dir_id in self.dir_freelist:

            stream_pos = dir_id * 128
            chain_index = stream_pos // self.sector_size
            sid_offset  = stream_pos % self.sector_size
            sid = self.dir_fat_chain[chain_index]

            pos = ((sid + 1) *  self.sector_size) + sid_offset

            f.seek(pos)
            f.write(empty_dir)

    def next_free_minifat_sect(self):

        idx_per_sect = self.sector_size // self.mini_stream_sector_size
        stream_sects = len(self.mini_stream_chain) * idx_per_sect

        if self.minifat_freelist:
            i = self.minifat_freelist.pop(0)
            assert self.minifat[i] == FREESECT
            if i+1 > stream_sects:
                self.mini_stream_grow()
            return i

        # if we got here need to add additional fat
        sid = self.next_free_sect()
        # logging.warning("growing minifat to sid %d" % sid)

        idx_start = len(self.minifat)
        idx_end = idx_start + self.sector_size // 4

        self.minifat.extend([FREESECT for i in range(idx_start, idx_end)])
        self.minifat_freelist.extend([i for i in range(idx_start, idx_end)])

        if self.minifat_sector_count == 0:
            self.minifat_sector_count = 1
            self.minifat_sector_start = sid
        else:
            self.minifat_sector_count += 1
            self.fat[self.minifat_chain[-1]] = sid

        self.minifat_chain.append(sid)
        self.fat[sid] = ENDOFCHAIN

        return self.next_free_minifat_sect()

    def next_free_sect(self):

        if self.fat_freelist:
            # print("using fat free list")
            i = self.fat_freelist.pop(0)
            assert self.fat[i] == FREESECT

            # Handle Range Lock Sector
            if i == RANGELOCKSECT and self.sector_size == 4096:
                self.fat[i] = ENDOFCHAIN
                logging.warning("range lock sector in fat freelist, marking ENDOFCHAIN")
                return self.next_free_sect()
            return i

        # if we got here need to add additional fat
        # logging.debug("fat full, growing")

        difat_table = None
        difat_index = None

        for t, i, v in self.iter_difat():
            if v == FREESECT:
                difat_table = t
                difat_index = i
                break

        new_difat_sect = None
        if difat_index is None:
            new_difat_sect = len(self.fat) + 1
            logging.debug("adding new difat to sid: %d" % new_difat_sect)
            if self.difat_sector_count == 0:
                self.difat_sector_start = new_difat_sect
                self.difat_sector_count = 1
            else:
                self.difat[-1][-1] = new_difat_sect
                self.difat_sector_count += 1

            # add difat table
            difat = []
            for i in range(self.sector_size // 4):
                difat.append(FREESECT)

            difat[-1] == ENDOFCHAIN
            self.difat.append(difat)

            for t, i, v in self.iter_difat():
                if v == FREESECT:
                    difat_table = t
                    difat_index = i
                    break

        new_fat_sect = len(self.fat)
        # logging.debug("adding new fat to sid: %d" % new_fat_sect)

        self.difat[difat_table][difat_index] = new_fat_sect

        # grow fat entries
        idx_start = len(self.fat)
        idx_end = idx_start + (self.sector_size // 4)

        self.fat.extend([FREESECT for i in range(self.sector_size // 4)])

        non_free_sids = set([new_fat_sect, new_difat_sect])

        # Handle Range Lock Sector
        # The range lock sector is the sector
        # that covers file offsets 0x7FFFFF00-0x7FFFFFFF in the file
        if RANGELOCKSECT < idx_end and RANGELOCKSECT > idx_start and self.sector_size == 4096:
            non_free_sids.add(RANGELOCKSECT)
            logging.debug("adding range lock")
            self.fat[RANGELOCKSECT] = ENDOFCHAIN

        freelist = [i for i in range(idx_start, idx_end) if i not in non_free_sids]

        self.fat_freelist.extend(freelist)

        self.fat[new_fat_sect] = FATSECT
        self.fat_sector_count += 1

        if not new_difat_sect is None:
            self.fat[new_difat_sect] = DIFSECT

        return self.next_free_sect()

    def read_sector_data(self, sid):

        sector_data = self.sector_cache.get(sid, None)
        if sector_data is not None:
            return sector_data
        else:
            pos = (sid + 1) *  self.sector_size
            self.f.seek(pos)
            sector_data = bytearray(self.sector_size)
            # NOTE: if requested sector doesn't exist or
            # is truncated will pad with zeros, expected behavior
            bytes_read = self.f.readinto(sector_data)
            self.sector_cache[sid] = sector_data
            return sector_data

    def get_sid_offset(self, abs_pos):
        sid, sid_offset = divmod(abs_pos, self.sector_size)
        return sid-1, sid_offset

    def dir_entry_sid_offset(self, dir_id):
        stream_pos = dir_id * 128
        chain_index, sid_offset = divmod(stream_pos, self.sector_size)
        sid = self.dir_fat_chain[chain_index]
        return sid, sid_offset

    def dir_entry_pos(self, dir_id):
        sid, sid_offset = self.dir_entry_sid_offset(dir_id)
        pos = ((sid + 1) *  self.sector_size) + sid_offset
        return pos

    def read_dir_entry(self, dir_id, parent = None):
        if dir_id is None:
            return None

        entry = self.dir_cache.get(dir_id, None)
        if entry is not None:
            return entry

        # assert not dir_id in self.dir_freelist

        stream_pos = dir_id * 128
        chain_index = stream_pos // self.sector_size
        sid_offset  = stream_pos % self.sector_size
        sid = self.dir_fat_chain[chain_index]

        sector_data = self.read_sector_data(sid)

        data= bytearray(sector_data[sid_offset:sid_offset+128])
        entry = DirEntry(self, dir_id, data=data)

        entry.parent = parent
        self.dir_cache[dir_id] = entry
        return entry

    def clear_sector(self, sid):
        sector_pos = (sid + 1) * self.sector_size
        self.f.seek(sector_pos)
        self.f.write(bytearray(self.sector_size))

    def next_free_dir_id(self):

        # use free list first
        if self.dir_freelist:
            return self.dir_freelist.pop(0)

        f = self.f

        sect = self.fat_chain_append(self.dir_fat_chain[-1])

        self.dir_fat_chain.append(sect)
        self.dir_sector_count += 1

        first_dir_id = (len(self.dir_fat_chain) - 1) * self.sector_size // 128
        last_dir_id = first_dir_id + (self.sector_size // 128)
        self.dir_freelist.extend(range(first_dir_id, last_dir_id))

        return self.next_free_dir_id()

    def get_fat_chain(self, start_sid, minifat=False):
        fat = self.fat
        fat_name = "FAT"
        if minifat:
            fat = self.minifat
            fat_name = "MINIFAT"

        # Floyd's Tortoise and Hare cycle-finding algorithm
        a = start_sid
        b = start_sid
        sectors = []

        if start_sid in (None, ENDOFCHAIN, FREESECT, DIFSECT, FATSECT):
            return []

        while b != ENDOFCHAIN:
            sectors.append(b)
            b = fat[b]
            if a != ENDOFCHAIN:
                a = fat[a]
                if a != ENDOFCHAIN:
                    a = fat[a]
                    if a == b:
                        raise CompoundFileBinaryError('cyclic %s fat chain found starting at %d' % (fat_name, start_sid))

        return sectors

    def mini_stream_grow(self):
        sid = self.next_free_sect()
        # logging.debug("adding to mini stream fat sid: %d" %  sid)
        if not self.mini_stream_chain:
            self.mini_stream_chain = [sid]
            self.root.sector_id = sid
        else:
            self.fat[self.mini_stream_chain[-1]] = sid
            self.mini_stream_chain.append(sid)

        self.fat[sid] = ENDOFCHAIN

    def fat_chain_append(self, start_sid, minifat=False):

        if minifat:
            sect = self.next_free_minifat_sect()
            # logging.debug("creating new mini sector: %d" % sect)
            fat = self.minifat
        else:
            sect = self.next_free_sect()
            # logging.debug("creating new sector: %d" % sect)
            fat = self.fat

        if start_sid is None:
            fat[sect] = ENDOFCHAIN
        else:
            fat_chain = self.get_fat_chain(start_sid, minifat)
            assert fat_chain
            fat[fat_chain[-1]] = sect
            fat[sect] = ENDOFCHAIN

        return sect

    def free_fat_chain(self, start_sid, minifat=False):
        fat =self.fat
        if minifat:
            fat = self.minifat

        for sid in self.get_fat_chain(start_sid, minifat):
            fat[sid] = FREESECT
            if minifat:
                self.minifat_freelist.insert(0, sid)
            else:
                self.fat_freelist.insert(0, sid)


    def create_dir_entry(self, path, dir_type='storage', class_id=None):

        if self.exists(path):
            raise ValueError("%s already exists" % path)

        dirname = os.path.dirname(path)
        basename = os.path.basename(path)

        root = self.find(dirname)

        if root is None:
            raise ValueError("parent dirname does not exist: %s" % dirname)

        if not root.type in ('storage', 'root storage'):
            raise ValueError("can not add entry to non storage type")

        dir_id = self.next_free_dir_id()
        logging.debug("next dir id %d" % dir_id)

        entry = DirEntry(self, dir_id)
        entry.name = basename
        entry.type = dir_type
        entry.class_id = class_id

        root.add_child(entry)
        self.dir_cache[dir_id] = entry

        return entry

    def free_dir_entry(self, entry):

        # add freelist
        self.dir_freelist.append(entry.dir_id)

        # remove from dir caches
        if entry.dir_id in self.dir_cache:
            del self.dir_cache[entry.dir_id]

        if entry.dir_id in self.children_cache:
            del self.children_cache[entry.dir_id]

        if entry.dir_id in self.modified:
            del self.modified[entry.dir_id]

        entry.dir_id = None


    def remove(self, path):
        """
        Removes both streams and storage DirEntry types from file.
        storage type entries need to be empty dirs.
        """

        entry = self.find(path)

        if not entry:
            raise ValueError("%s does not exists" % path)

        if entry.type == 'root storage':
            raise ValueError("can no remove root entry")

        if entry.type == "storage" and not entry.child_id is None:
            raise ValueError("storage contains children")

        entry.pop()

        # remove stream data
        if entry.type == "stream":
            self.free_fat_chain(entry.sector_id, entry.byte_size < self.min_stream_max_size)

        self.free_dir_entry(entry)


    def rmtree(self, path):
        """
        Removes directory structure, similar to shutil.rmtree.
        """
        for root, storage, streams in self.walk(path, topdown=False):

            for item in streams:
                self.free_fat_chain(item.sector_id, item.byte_size < self.min_stream_max_size)
                self.free_dir_entry(item)

            for item in storage:
                self.free_dir_entry(item)

            root.child_id = None

        # remove root item
        self.remove(path)


    def listdir(self, path = None):
        """
        Return a list containing the ``DirEntry`` objects in the directory
        given by path.
        """

        result = self.listdir_dict(path)
        return result.values()

    def listdir_dict(self, path = None):
        """
        Return a dict containing the ``DirEntry`` objects in the directory
        given by path with name of the dir as key.
        """

        if path is None:
            path = self.root

        root = self.find(path)
        if root is None:
            raise ValueError("unable to find dir: %s" % str(path))

        if not root.isdir():
            raise ValueError("can only list storage types")

        children = self.children_cache.get(root.dir_id, None)
        if children is not None:
            return children

        child = root.child()

        result = {}
        if not child:
            self.children_cache[root.dir_id] = result
            return result

        dir_per_sector = self.sector_size // 128
        max_dirs_entries = self.dir_sector_count * dir_per_sector

        stack =  deque([child])
        count = 0

        while stack:
            current = stack.pop()
            result[current.name] = current
            count += 1

            if count > max_dirs_entries:
                raise CompoundFileBinaryError("corrupt folder structure")

            left = current.left()
            if left:
                stack.append(left)
            right =  current.right()
            if right:
                stack.append(right)

        self.children_cache[root.dir_id] = result
        return result

    def find(self, path):
        """
        find a ``DirEntry`` located at *path*. Returns ``None`` if path
        does not exist.
        """

        if isinstance(path, DirEntry):
            return path

        if path == "/":
            return self.root

        split_path = path.lstrip('/').split("/")

        i = 0
        root = self.root

        while True:

            children = self.listdir_dict(root)
            match = children.get(split_path[i], None)

            if match:
                if i == len(split_path) - 1:
                    return match
                root = match
                i += 1
            else:
                return None

    def walk(self, path = None, topdown=True):
        """
        Similar to :func:`os.walk`, yeields a 3-tuple ``(root, storage_items, stream_items)``
        """

        if path is None:
            path = self.root

        root = self.find(path)

        if not root.isdir():
            raise ValueError("can only walk storage types")

        if not root.child_id:
            return

        if topdown:
            storage_items = []
            stream_items = []

            for item in self.listdir(root):
                if item.isdir():
                    storage_items.append(item)
                else:
                    stream_items.append(item)

            yield root, storage_items, stream_items

            for item in storage_items:
                for root, storage_items, stream_items in self.walk(item):
                    yield root, storage_items, stream_items
        else:

            def topdown_visit_node(root):
                storage_items = []
                stream_items = []
                for item in self.listdir(root):
                    if item.isdir():
                        for sub_root, sub_storage, sub_stream in topdown_visit_node(item):
                            yield sub_root, sub_storage, sub_stream

                        storage_items.append(item)
                    else:
                        stream_items.append(item)

                yield root, storage_items, stream_items

            for root_item, storage, stream in topdown_visit_node(root):
                yield root_item, storage, stream

    def validate_directory_structure(self):
        for root, storage, stream in self.walk():
            validate_rbtree(root.child())


    def exists(self, path):
        """
        Return ``True`` if path refers to a existing path.
        """
        if self.find(path) is None:
            return False
        return True

    def makedir(self, path, class_id=None):
        """
        Create a storage DirEntry name path
        """
        return self.create_dir_entry(path, dir_type='storage', class_id=class_id)

    def makedirs(self, path):
        """
        Recursive storage DirEntry creation function.
        """
        root = ""

        assert path.startswith('/')
        p = path.strip('/')
        for item in p.split('/'):
            root += "/" + item
            if not self.exists(root):
                self.makedir(root)

        return self.find(path)

    def move(self, src, dst):
        """
        Moves ``DirEntry`` from src to dst
        """
        src_entry = self.find(src)
        if src_entry is None:
            raise ValueError("src path does not exist: %s" % src)

        if dst.endswith('/'):
            dst += src_entry.name

        if self.exists(dst):
            raise ValueError("dst path already exist: %s" % dst)

        if dst == '/' or src == '/':
            raise ValueError("cannot overwrite root dir")

        split_path = dst.strip('/').split('/')
        dst_basename = split_path[-1]
        dst_dirname = '/' + '/'.join(split_path[:-1])

        # print(dst)
        # print(dst_basename, dst_dirname)

        dst_entry = self.find(dst_dirname)
        if dst_entry is None:
            raise ValueError("src path does not exist: %s" % dst_dirname)

        if not dst_entry.isdir():
            raise ValueError("dst dirname cannot be stream: %s" % dst_dirname)

        # src_entry.parent.remove_child(src_entry)

        src_entry.pop()

        src_entry.parent = None
        src_entry.name = dst_basename
        dst_entry.add_child(src_entry)

        self.children_cache[dst_entry.dir_id][src_entry.name] = src_entry

        return src_entry

    def open(self, path, mode='r'):
        """Open stream, returning ``Stream`` object"""

        entry = self.find(path)
        if entry is None:
            if mode == 'r':
                raise ValueError("stream does not exists: %s" % path)
            entry = self.create_dir_entry(path, 'stream', None)

        else:
            if not entry.isfile():
                raise ValueError("can only open stream type DirEntry's")

            if mode == 'w':
                logging.debug("stream: %s exists, overwriting" % path)
                self.free_fat_chain(entry.sector_id, entry.byte_size < self.min_stream_max_size)
                entry.sector_id = None
                entry.byte_size = 0
                entry.class_id = None
            elif mode == 'rw':
                pass

        s = Stream(self, entry, mode)
        return s



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
)
import os
import unittest
import common
import aaf2
from aaf2 import ama

avc_profiles = [('yuv420p', 'baseline'),
                ('yuv420p', 'main'),
                ('yuv420p', 'high'),
                ('yuv422p', 'high422'),
                # ('yuv444p', 'high444')# unsupported in MC?
                ]

prores_profiles = [
    (0, 'proxy'),
    (1, 'LT'),
    (2, 'standard'),
    (3, 'HQ'),
    (4, '4444'),
]


class AMATests(unittest.TestCase):


    def assert_mastermob_valid_edit_spec(self, master_mob, expected_picture_slots, expected_sound_slots):
        """
        Verifies that a MasterMob has all the required elements for the edit spec
        :param master_mob:
        :return:
        """
        physical_sound_tracks = []
        self.assertIsNotNone(master_mob, "Failed to find MasterMob by saved ID")
        self.assertEqual(len(master_mob.slots), expected_picture_slots + expected_sound_slots,
                         "Failed to find correct number of slots in MasterMob")

        sound_slots = list([slot for slot in master_mob.slots if slot.segment.media_kind == 'Sound'])
        picture_slots = list([slot for slot in master_mob.slots if slot.segment.media_kind == 'Picture'])

        self.assertEqual(len(sound_slots), expected_sound_slots, "MasterMob has incorrect number of sound slots")
        self.assertEqual(len(picture_slots), expected_picture_slots, "MasterMon has incorrect number of picture slots")

        file_source_mob = None
        for mob_slot in master_mob.slots:
            self.assertIsInstance(mob_slot.segment, aaf2.components.Sequence, "MobSlot has no Sequence")
            self.assertEqual(len(mob_slot.segment.components), 1, "Mob slot must have exactly one component")
            self.assertIsInstance(mob_slot.segment.components[0], aaf2.components.SourceClip,
                                  "Mob slot sequence has incorrect component type")
            source_clip = mob_slot.segment.components[0]

            self.assertIsNotNone(source_clip.start, "SourceClip does not have a start time")
            self.assertIsNotNone(source_clip.length, "SourceClip does not have a length time")
            self.assertTrue(source_clip.length > 0, "SourceClip appears to have invalid length")

            self.assertIsNotNone(source_clip.mob, "MasterMob SourceClip does not have a linked SourceMob")
            if file_source_mob is None:
                file_source_mob = source_clip.mob
            else:
                self.assertEqual(source_clip.mob, file_source_mob,
                                 "MasterMob slots appear to reference different SourceMOBs")

            if mob_slot.media_kind == 'Sound':
                physical_sound_tracks.append(int(mob_slot['PhysicalTrackNumber'].value))

            self.assertEqual(len(source_clip.mob.slots), expected_picture_slots + expected_sound_slots,
                             "SourceMob has incorrect number of slots")

            tape_source_clip = file_source_mob.slot_at(mob_slot.slot_id).segment.components[0]
            self.assertIsInstance(tape_source_clip, aaf2.components.SourceClip,
                                  "File SourceMob does not have a tape SourceMob")

            if isinstance(tape_source_clip.mob.descriptor, aaf2.essence.TapeDescriptor):
                pass
            elif isinstance(tape_source_clip.mob.descriptor, aaf2.essence.ImportDescriptor):
                pass
            else:
                self.fail("Tape SourceMob descriptor must be either a TapeDescriptor or ImportDescriptor")

        for chan_num in range(expected_sound_slots):
            self.assertEqual(physical_sound_tracks.count(chan_num + 1), 1,
                             "Incorrect PhysicalTrackNumber property on master mob slot")

        self.assertEqual(len(physical_sound_tracks), expected_sound_slots,
                         "Incorrect PhysicalTrackNumber count on master mob slot")

    def assert_valid_multiple_descriptor(self, mastermob, expected_audio_channel_count):
        for mob_slot in mastermob.slots:
            source_clip = mob_slot.segment.components[0]
            self.assertIsNotNone(source_clip.mob.comments['Video'],
                                 "SourceMob must have a value for 'Video' comment")

            self.assertEqual(len(source_clip.mob.descriptor['FileDescriptors'].value), 2,
                             "SourceMob's descriptor has incorrect 'FileDescriptor' property value")

            self.assertIsInstance(source_clip.mob.descriptor, aaf2.essence.MultipleDescriptor,
                                  "SourceClip Mob has incorrect descriptor")

            for descriptor in source_clip.mob.descriptor['FileDescriptors'].value:
                self.assertIsNotNone(descriptor['Locator'].value,
                                     "SourceClip descriptor not properly formatted")

                locators = descriptor['Locator'].value
                self.assertTrue(len(locators) >= 1)

                if isinstance(descriptor, aaf2.essence.PCMDescriptor):
                    self.assertEqual(descriptor['Channels'].value, expected_audio_channel_count,
                                     "SourceClip descriptor not properly formatted")
                elif isinstance(descriptor, aaf2.essence.CDCIDescriptor):
                    self.assertIsInstance(descriptor['ComponentWidth'].value, int)
                    self.assertIsInstance(descriptor['HorizontalSubsampling'].value, int)

                else:
                    self.fail("Encountered unexpected essence descriptor")

    def test_monoaural_wav(self):
        new_file = os.path.join(common.sandbox(), 'ama_wav.aaf')
        with aaf2.open(new_file, 'w') as f:
            wavfile = common.generate_pcm_audio_mono('test_ama.wav', fmt='wav')
            meta = common.probe(wavfile)
            mobs = ama.create_media_link(f, wavfile, meta)
            self.assertTrue(len(mobs), 3)

        with aaf2.open(new_file, 'r') as f:
            common.walk_aaf(f.root)
            self.assertTrue(len(f.content.mobs) == 3)
            self.assertTrue(len(list(f.content.mastermobs())) == 1)
            master_mob = next(f.content.mastermobs())
            self.assert_mastermob_valid_edit_spec(master_mob=master_mob, expected_sound_slots=1,
                                                  expected_picture_slots=0)
            self.assertEqual(len(master_mob.slots), 1, "MasterMob should only have one slot")
            self.assertEqual(master_mob.slots[0].media_kind, 'Sound', "MasterMob slot has incorrect media_kind")
            source_clip = master_mob.slots[0].segment.components[0]
            descriptor = source_clip.mob.descriptor
            self.assertIsNotNone(descriptor, "File SourceMob has no WAVEDescriptor")
            self.assertIsInstance(descriptor, aaf2.essence.WAVEDescriptor, "File SourceMob has no WAVEDescriptor")
            self.assertIsNotNone(descriptor['Summary'].value, "WAVEDescriptor missing required 'Summary' property")

    def test_monoaural_aiff(self):
        new_file = os.path.join(common.sandbox(), 'ama_aiff.aaf')
        with aaf2.open(new_file, 'w') as f:
            aiff_file = common.generate_pcm_audio_mono('test_ama_aiff', fmt='aiff')
            meta = common.probe(aiff_file)
            mobs = ama.create_media_link(f, aiff_file, meta)
            self.assertTrue( len(mobs), 3)

        with aaf2.open(new_file, 'r') as f:
            common.walk_aaf(f.root)
            self.assertTrue(len(f.content.mobs) == 3)
            self.assertTrue(len(list(f.content.mastermobs())) == 1)
            master_mob = next(f.content.mastermobs())
            self.assert_mastermob_valid_edit_spec(master_mob=master_mob, expected_sound_slots=1,
                                                  expected_picture_slots=0)
            self.assertEqual(len(master_mob.slots), 1, "MasterMob should only have one slot")
            self.assertEqual(master_mob.slots[0].media_kind, 'Sound', "MasterMob slot has incorrect media_kind")
            source_clip = master_mob.slots[0].segment.components[0]
            descriptor = source_clip.mob.descriptor
            self.assertIsNotNone(descriptor, "File SourceMob has no WAVEDescriptor")
            self.assertIsInstance(descriptor, aaf2.essence.AIFCDescriptor, "File SourceMob has no AIFCDescriptor")
            self.assertIsNotNone(descriptor['Summary'].value, "AIFCDescriptor missing required 'Summary' property")

    def test_avc_mov(self):
        new_file = os.path.join(common.sandbox(), 'avc_mov.aaf')

        audio_channel_count = 2
        created_mastermob_ids = []
        with aaf2.open(new_file, 'w') as f:
            for (pix_fmt, profile) in avc_profiles:
                vcodec = ['-pix_fmt', pix_fmt, '-c:v', 'h264', '-profile:v', profile]

                mov = common.generate_mov('ama_avc_%s.mov' % profile, overwrite=False, vcodec=vcodec,
                                          audio_channels=audio_channel_count)
                meta = common.probe(mov)
                # print(meta['streams'][0]['profile'])

                mobs = f.content.create_ama_link(mov, meta)
                self.assertEqual(len(mobs), 3)
                self.assertIsInstance(mobs[0], aaf2.mobs.MasterMob)
                self.assertIsInstance(mobs[1], aaf2.mobs.SourceMob)
                self.assertIsInstance(mobs[2], aaf2.mobs.SourceMob)
                created_mastermob_ids.append(mobs[0].mob_id)

        with aaf2.open(new_file, 'r') as f:
            common.walk_aaf(f.root)
            self.assertEqual(len(f.content.mobs), len(avc_profiles) * 3,
                             "Failed to create exactly three MOBs per avc_profile")
            self.assertEqual(len(list(f.content.mastermobs())), len(avc_profiles),
                             "Failed to create exactly one MasterMOB per avc_profile")

            for mastermob_id in created_mastermob_ids:
                mastermob = next((mob for mob in f.content.mobs if mob.mob_id == mastermob_id), None)
                self.assert_mastermob_valid_edit_spec(mastermob, expected_picture_slots=1, expected_sound_slots=2)
                self.assert_valid_multiple_descriptor(mastermob, audio_channel_count)

    def test_prores(self):
        new_file = os.path.join(common.sandbox(), 'prores_mov.aaf')
        created_mastermob_ids = []
        with aaf2.open(new_file, 'w') as f:

            for profile, name in prores_profiles:
                vcodec = ['-c:v', 'prores_ks', '-profile:v', str(profile)]
                mov = common.generate_mov('ama_prores_%s.mov' % (name,), overwrite=False, vcodec=vcodec,
                                          audio_channels=2)
                meta = common.probe(mov)
                mobs = ama.create_media_link(f, mov, meta)
                self.assertEqual(len(mobs), 3, "create_ama_link must return exactly three mobs")
                created_mastermob_ids.append(mobs[0].mob_id)

        with aaf2.open(new_file, 'r') as f:
            common.walk_aaf(f.root)
            self.assertEqual(len(f.content.mobs), len(prores_profiles) * 3,
                             "Failed to create exactly three MOBs per prores_profile")
            self.assertEqual(len(list(f.content.mastermobs())), len(prores_profiles),
                             "Failed to create exactly one MasterMOB per prores_profile")

            for mastermob_id in created_mastermob_ids:
                mastermob = next((mob for mob in f.content.mobs if mob.mob_id == mastermob_id), None)
                self.assert_mastermob_valid_edit_spec(mastermob, expected_picture_slots=1, expected_sound_slots=2)
                self.assert_valid_multiple_descriptor(mastermob, 2)


if __name__ == "__main__":
    import logging

    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
import os
import subprocess
import unittest
import hashlib
import common
import aaf2
from uuid import UUID
from aaf2 import video, audio, mobid, mobs, essence


class WAVETests(unittest.TestCase):
    
    def test_wav_link(self):
        test_outfile = os.path.join(common.sandbox(), 'wav_link.aaf')

        test_link_targets = [
            common.generate_pcm_audio_mono('testmono', sample_rate=48000, duration=10,
                sample_format='pcm_s16le', fmt='wav'),
            common.generate_pcm_audio_mono('testmono24', sample_rate=48000, duration=10,
                sample_format='pcm_s24le', fmt='wav'),
            common.generate_pcm_audio_stereo('testmono', sample_rate=48000, duration=10,
                sample_format='pcm_s24le', fmt='wav')
                ]

        with aaf2.open(test_outfile,'w') as f:
            for target in test_link_targets:
                meta = common.probe(target)
                mobs = f.content.link_external_wav(meta)
                

if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
from datetime import datetime
import io

from . import core
from . utils import register_class
from .auid import AUID
from .rational import AAFRational
from .interpolation import integrate_iter, lerp, cubic_interpolate, bezier_interpolate


ConstantInterp     = AUID("5b6c85a5-0ede-11d3-80a9-006008143e6f")
LinearInterp       = AUID("5b6c85a4-0ede-11d3-80a9-006008143e6f")
BezierInterpolator = AUID("df394eda-6ac6-4566-8dbe-f28b0bdd781a")
CubicInterpolator  = AUID("a04a5439-8a0e-4cb7-975f-a5b255866883")

class TaggedValueHelper(object):
    def __init__(self, poperty_vector):
        self.p = poperty_vector

    def get(self, key, default=None):
        for item in self.p:
            if item['Name'].value == key:
                return item
        return default

    def __contains__(self, key):
        return not self.get(key, None) is None

    def __getitem__(self, key):
        p = self.get(key, None)
        if p:
            return p['Value'].value

        raise IndexError(key)

    def items(self):
        for item in self.p:
            yield item['Name'].value, item["Value"].value

    def append(self, value):
        self.p.append(value)

    def __setitem__(self, key, value):
        tag = self.get(key, None)
        if tag is None:
            tag = self.p.parent.root.create.TaggedValue()
            tag['Name'].value = key
            self.p.append(tag)

        tag['Value'].value = value

@register_class
class TaggedValue(core.AAFObject):
    class_id = AUID("0d010101-0101-3f00-060e-2b3402060101")
    __slots__ = ()

    def __init__(self, name=None, value=None, value_typedef=None):
        self.name = name
        if value is not None:
            self.encode_value(value, value_typedef)

    @property
    def name(self):
        return self['Name'].value

    @name.setter
    def name(self, value):
        self['Name'].value = value

    @property
    def value(self):
        return self['Value'].value

    @value.setter
    def value(self, value):
        self['Value'].value = value

    @property
    def value_typedef(self):
        if self['Value'].data:
            return self['Value'].typedef.decode_typedef(self['Value'].data)

    def encode_value(self, value, value_typedef=None):
        if value_typedef is None:
            self.value = value
            return
        self['Value'].add_pid_entry()
        self['Value'].data = self['Value'].typedef.encode(value, value_typedef)
        self['Value'].mark_modified()

    def __repr__(self):
        s = "%s.%s" % (self.__class__.__module__,
                                self.__class__.__name__)

        value_typedef = self.value_typedef
        if value_typedef:
            s += ' %s' % repr(value_typedef)

        name = self.name
        if name:
            s += ' %s' % name

        s += " = " + repr(self.value)

        return '<%s at 0x%x>' % (s, id(self))

class Parameter(core.AAFObject):
    class_id = AUID("0d010101-0101-3c00-060e-2b3402060101")
    __slots__ = ()

    @property
    def auid(self):
        return self['Definition'].value
    @auid.setter
    def auid(self, value):
        self['Definition'].value = value

    @property
    def parameterdef(self):
        return self.root.dictionary.lookup_parameterdef(self['Definition'].value)

    @parameterdef.setter
    def parameterdef(self, value):
        self['Definition'].value = value.auid

    @property
    def name(self):
        return self.parameterdef.name

    @property
    def unique_property(self):
        return self['Definition']

    @property
    def unique_key(self):
        return self.unique_property.value

    def __repr__(self):
        s = "%s.%s" % (self.__class__.__module__,
                                self.__class__.__name__)

        parameterdef = self.parameterdef
        if parameterdef:
            s += ' %s' % parameterdef.name

        return '<%s at 0x%x>' % (s, id(self))

@register_class
class ConstantValue(Parameter):
    class_id = AUID("0d010101-0101-3d00-060e-2b3402060101")
    __slots__ = ()
    def __init__(self, parameterdef=None, value=None):
        super(ConstantValue, self).__init__()

        if parameterdef is not None:
            self.parameterdef = self.root.dictionary.lookup_parameterdef(parameterdef)

        if value is not None:
            if parameterdef is None:
                raise ValueError("need parameterdef to initialize value")
            self.value = value

    @property
    def typedef(self):
        return self.parameterdef.typedef

    def value_at(self, t):
        return self.value

    @property
    def value(self):
        return self['Value'].value

    @value.setter
    def value(self, value):
        self['Value'].add_pid_entry()
        indirect_typedef = self['Value'].typedef
        parameter_typdef = self.typedef
        self['Value'].data = indirect_typedef.encode(value, parameter_typdef)
        self['Value'].mark_modified()


@register_class
class VaryingValue(Parameter):
    class_id = AUID("0d010101-0101-3e00-060e-2b3402060101")
    __slots__ = ()

    def __init__(self, parameterdef=None, interperlationdef=None):
        super(VaryingValue, self).__init__()
        if parameterdef:
            self.parameterdef = self.root.dictionary.lookup_parameterdef(parameterdef)

        if interperlationdef:
            self.interpolationdef = self.root.dictionary.lookup_interperlationdef(interperlationdef)

    @property
    def interpolationdef(self):
        return self['Interpolation'].value

    @interpolationdef.setter
    def interpolationdef(self, value):
        self['Interpolation'].value = value

    # TODO: Deprecate this
    @property
    def interpolation(self):
        return self['Interpolation'].value

    @interpolation.setter
    def interpolation(self, value):
        self['Interpolation'].value = value

    @property
    def pointlist(self):
        return self['PointList']

    @property
    def typedef(self):
        return self.parameterdef.typedef

    def add_keyframe(self, time, value, edit_hint=None):
        cp = self.root.create.ControlPoint()
        cp.time = time

        cp['Value'].add_pid_entry()
        cp['Value'].data = cp['Value'].typedef.encode(value, self.typedef)
        cp['Value'].mark_modified()

        if edit_hint:
            cp['EditHint'].value = edit_hint

        pointlist = self['PointList']

        if len(pointlist) > 0:
            index = self.nearest_index(time)
            nearest = pointlist[index]
            if float(nearest.time) == float(time):
                pointlist[index] = cp
            else:
                pointlist.insert(index+1, cp)
        else:
            pointlist.append(cp)

        return cp

    def value_at(self, t):
        t = float(t)

        index = self.nearest_index(t)
        pointlist = self['PointList']

        p1 = pointlist[index]

        # clamp if t outside of range
        if t < p1.time or index + 1 >= len(pointlist):
            return float(p1.value)

        if self.interpolationdef.auid == ConstantInterp:
            return float(p1.value)

        p2 = pointlist[index+1]

        if self.interpolationdef.auid == LinearInterp:
            t_len = float(p2.time) - float(p1.time)
            t_diff = t - float(p1.time)
            t_mix = t_diff/t_len

            v0 = float(p1.value)
            v1 = float(p2.value)
            return lerp(v0, v1, t_mix)

        elif self.interpolationdef.auid == BezierInterpolator:
            t0 = float(p1.time)
            v0 = float(p1.value)

            t3 = float(p2.time)
            v3 = float(p2.value)

            tangents = p1.tangents[1]
            t1 = t0 + tangents[0]
            v1 = v0 + tangents[1]

            tangents = p2.tangents[0]
            t2 = t3 + tangents[0]
            v2 = v3 + tangents[1]

            return bezier_interpolate((t0, v0),
                                      (t1, v1),
                                      (t2, v2),
                                      (t3, v3), t)

        elif self.interpolationdef.auid == CubicInterpolator:

            t1 = float(p1.time)
            v1 = float(p1.value)

            t2 = float(p2.time)
            v2 = float(p2.value)

            if index - 1 >= 0:
                p0 = pointlist[index - 1]
                t0 = float(p0.time)
                v0 = float(p0.value)
            else:
                t0 = t1 - ((t2 - t1) * 0.5)
                v0 = v1

            if index + 2 < len(pointlist):
                p3 = pointlist[index + 2]
                t3 = float(p3.time)
                v3 = float(p3.value)
            else:
                t3 = t2 + ((t2 - t1) * 0.5)
                v3 = v2

            return cubic_interpolate((t0, v0),
                                     (t1, v1),
                                     (t2, v2),
                                     (t3, v3), t)

        else:

            raise NotImplementedError("Interpolation not implemented for %s %s" %
                           (self.interpolationdef.name, str(self.interpolationdef.auid)))

    def nearest_index(self, t):
        """
        binary search for index of point.time <= t
        """
        pointlist = self['PointList']
        start = 0
        end = len(pointlist) - 1
        while True:
            if end < start:
                return max(0, end)

            m = (start + end) // 2
            p = pointlist[m]
            if p.time < t:
                start = m + 1
            elif p.time > t:
                end = m - 1
            else:
                return m

@register_class
class ControlPoint(core.AAFObject):
    class_id = AUID("0d010101-0101-1900-060e-2b3402060101")
    __slots__ = ()

    @property
    def time(self):
        return float(self['Time'].value)

    @time.setter
    def time(self, value):
        self['Time'].value = value

    @property
    def value(self):
        return float(self['Value'].value)

    @value.setter
    def value(self, value):
        self['Value'].value = AAFRational(value)

    @property
    def point_properties(self):
        props = {}
        if 'ControlPointPointProperties' in self:
            for p in self['ControlPointPointProperties'].value:
                props[p.name] = p.value
        return props

    @property
    def base_frame(self):
        return self.point_properties.get("PP_BASE_FRAME_U", 0)

    @property
    def tangents(self):
        props = self.point_properties
        return [(float(props.get("PP_IN_TANGENT_POS_U", 0)),
                 float(props.get("PP_IN_TANGENT_VAL_U", 0))),
                (float(props.get("PP_OUT_TANGENT_POS_U", 0)),
                 float(props.get("PP_OUT_TANGENT_VAL_U", 0)))]

def generate_offset_map(speed_map, start=0, end=None):
    pointlist = speed_map['PointList']

    # first speed map key frame is the zero point
    # of the offset map curve
    first = pointlist.value[0]
    center = int(first.time)
    if end is None:
        last = pointlist.value[-1]
        end = int(last.time)

    if start > end:
        raise ValueError("start needs to be less then end")

    time = []
    value = []
    offset_index = None

    inter_start = min(start, center)
    inter_end = max(center,  end+1)

    for i, (t,v) in enumerate(integrate_iter(speed_map.value_at, inter_start, inter_end)):
        time.append(t)
        value.append(v)

        if t == center:
            offset_index = i

    center_offset = value[offset_index]

    # not really sure what this base frame offset is about
    # but appears to contain the how much calculation is off...
    center_offset -= first.base_frame

    result = []
    for i, t in enumerate(time):
        if t > end:
            break

        if t >= start:
            v = value[i] - center_offset
            result.append((t, v))

    return result



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
)

import os
import sys
from .rational import AAFRational
from . import video, audio, mxf
from .auid import AUID

import struct

MediaContainerGUID = {
    "Generic": (AUID("b22697a2-3442-44e8-bb8f-7a1cd290ebf1"),
                ('.3g2', '.3gp', '.aac', '.au', '.avi', '.bmp', '.dv', '.gif',
                 '.jfif', '.jpeg', '.jpg', '.m4a', '.mid', '.moov', '.mov',
                 '.movie', '.mp2', '.mp3', '.mp4', '.mpa', '.mpe', '.mpeg',
                 '.mpg', '.png', '.psd', '.qt', '.tif', '.tiff',)),
    "AVCHD": (AUID("f37d624b307d4ef59bebc539046cad54"),
              ('.mts', '.m2ts',)),
    "ImageSequencer": (AUID("4964178d-b3d5-485f-8e98-beb89d92a5f4"),
                       ('.dpx',)),
    "CanonRaw": (AUID("0f299461-ee19-459f-8ae6-93e65c76a892"),
                 ('.rmf',)),
    "WaveAiff": (AUID("3711d3cc-62d0-49d7-b0ae-c118101d1a16"),
                 ('.wav', '.wave', '.bwf', '.aif', '.aiff', '.aifc', '.cdda',)),
    "MXF": (AUID("60eb8921-2a02-4406-891c-d9b6a6ae0645"),
            ('.mxf',)),
    "QuickTime": (AUID("781f84b7-b989-4534-8a07-c595cb9a6fb8"),
                  ('.mov', '.mp4', '.m4v', '.mpg', '.mpe', '.mpeg', '.3gp', '.3g2',
                   '.qt', '.moov', '.movie', '.avi', '.mp2', '.mp3', '.m4a', '.wav',
                   '.aiff', '.aif', '.au', '.aac', '.mid', '.mpa', '.gif', '.jpg',
                   '.jpeg', '.jfif', '.tif', '.tiff', '.png', '.bmp', '.psd', '.dv')),
}


def get_wave_fmt(path):
    """
    Returns a bytearray of the WAVE RIFF header and fmt
    chunk for a `WAVEDescriptor` `Summary`
    """
    with open(path, 'rb') as file:
        if file.read(4) != b"RIFF":
            return None
        data_size = file.read(4)  # container size
        if file.read(4) != b"WAVE":
            return None
        while True:
            chunkid = file.read(4)
            sizebuf = file.read(4)
            if len(sizebuf) < 4 or len(chunkid) < 4:
                return None
            size = struct.unpack(b'<L', sizebuf)[0]
            if chunkid[0:3] != b"fmt":
                if size % 2 == 1:
                    seek = size + 1
                else:
                    seek = size
                file.seek(seek, 1)
            else:
                return bytearray(b"RIFF" + data_size + b"WAVE" + chunkid + sizebuf + file.read(size))

def get_aifc_fmt(path):
    """
    Compute the AIFC header information for a `AIFCDescriptor` `Summary`

    :param path: file to read chunk from
    :return: a `bytearray`
    """
    with open(path, 'rb') as file:
        if file.read(4) != b"FORM":
            return None
        data_size = file.read(4)
        signature = file.read(4)
        if signature not in (b"AIFF", b"AIFC"):
            return None

        while True:
            chunkid = file.read(4)
            sizebuf = file.read(4)
            if len(sizebuf) < 4 or len(chunkid) < 4:
                return None
            size = struct.unpack(">L", sizebuf)[0]
            if chunkid != b"COMM":
                if size % 2 == 1:
                    seek = size + 1
                else:
                    seek = size
                file.seek(seek, 1)
            else:
                return bytearray(b"FORM" + data_size + signature + chunkid + sizebuf + file.read(size))


def create_network_locator(f, absolute_path):
    n = f.create.NetworkLocator()
    if sys.version_info[0] < 3:
        import urllib
        n['URLString'].value = 'file://' + urllib.pathname2url(absolute_path)
    else:
        import pathlib
        n['URLString'].value = pathlib.Path(absolute_path).as_uri()

    return n

class FormatInfo:
    """
    Provides convenient access to commonly-used datums
    """

    def __init__(self, metadata):
        self.metadata = metadata

    @property
    def streams(self):
        for stream in self.metadata['streams']:
            yield StreamInfo(stream)

    @property
    def first_sound_stream(self):
        return next((stream for stream in self.streams if stream.is_sound), None)


    @property
    def first_picture_stream(self):
        return next((stream for stream in self.streams if stream.is_picture), None)

    def create_descriptor(self, f, path):
        if self.metadata['format']['format_name'] in ('wav',):
            return self.create_wav_descriptor(f, path)

        if self.metadata['format']['format_name'] in ('aiff',):
            return self.create_aifc_descriptor(f,path)

        elif self.metadata['format']['format_long_name'] == 'QuickTime / MOV':
            return self.create_multistream_descriptor(f, path)
        else:
            return None

    @property
    def container_guid(self):
        if self.metadata['format']['format_name'] in ('wav',):
            return MediaContainerGUID['WaveAiff'][0]

        # if self.metadata['format']['format_long_name'] == 'QuickTime / MOV':
        #     return MediaContainerGUID['QuickTime'][0]

        # just using the generic appears to work
        return MediaContainerGUID['Generic'][0]

    @property
    def edit_rate(self):
        """
        :return: The edit rate of the first picture stream, or if there are none, the first sound stream.
        """
        pix = self.first_picture_stream
        if pix is None:
            return self.first_sound_stream.edit_rate
        else:
            return pix.edit_rate

    @property
    def length(self):
        """
        :return: The length of the first picture stream, or if there are none, the first sound stream.
        """
        pix = self.first_picture_stream
        if pix is None:
            return self.first_sound_stream.length
        else:
            return pix.length

    def create_wav_descriptor(self, f, path):
        d = f.create.WAVEDescriptor()
        stream = self.first_sound_stream
        d['SampleRate'].value = stream.edit_rate
        d['Summary'].value = get_wave_fmt(path)
        d['Length'].value = stream.length
        d['ContainerFormat'].value = f.dictionary.lookup_containerdef("AAF")
        d['Locator'].append(create_network_locator(f, path))
        return d

    def create_aifc_descriptor(self, f, path):
        d = f.create.AIFCDescriptor()
        stream = self.first_sound_stream
        d['SampleRate'].value = stream.edit_rate
        d['Summary'].value = get_aifc_fmt(path)
        d['Length'].value = stream.length
        d['ContainerFormat'].value = f.dictionary.lookup_containerdef("AAF")
        d['Locator'].append(create_network_locator(f, path))
        return d

    def coalesce_descriptors(self, f, descriptors, path):
        if len(descriptors) > 1:
            desc = f.create.MultipleDescriptor()
            desc['Length'].value = self.length
            desc['SampleRate'].value = self.edit_rate
            desc['MediaContainerGUID'].value = self.container_guid
            desc['Locator'].append(create_network_locator(f, path))
            desc['FileDescriptors'].value = descriptors
            return desc
        else:
            return descriptors[0]

    def create_multistream_descriptor(self, f, path):
        descriptor_list = []

        for stream in self.streams:
            if stream.is_picture:
                desc = stream.create_video_descriptor(f)
                descriptor_list.append(desc)
                desc['Locator'].append(create_network_locator(f, path))

            elif stream.is_sound:
                desc = stream.create_pcm_descriptor(f)
                descriptor_list.append(desc)
                desc['Locator'].append(create_network_locator(f, path))

        return self.coalesce_descriptors(f, descriptor_list, path)


class StreamInfo:
    def __init__(self, metadata):
        self.metadata = metadata

    @property
    def codec_type(self):
        return self.metadata['codec_type']

    @property
    def codec_name(self):
        return self.metadata['codec_name']

    @property
    def is_sound(self):
        return self.codec_type == 'audio'

    @property
    def is_picture(self):
        return self.codec_type == 'video'

    @property
    def edit_rate(self):
        if self.is_sound:
            return AAFRational(self.metadata['sample_rate'])
        elif self.is_picture:
            return AAFRational(self.metadata['avg_frame_rate'])

    @property
    def length(self):
        if self.is_sound:
            return int(self.metadata['duration_ts'])
        elif self.is_picture:
            return int(self.metadata['nb_frames'])

    @property
    def physical_track_count(self):
        if self.is_sound:
            return self.metadata['channels']

    def create_pcm_descriptor(self, f):
        if not self.is_sound:
            return None

        d = f.create.PCMDescriptor()
        d['SampleRate'].value = self.edit_rate
        d['AudioSamplingRate'].value = self.edit_rate

        d['Channels'].value = self.physical_track_count
        d['AverageBPS'].value = int(self.metadata['bit_rate'])

        bit_depth, block_align = audio.audio_format_sizes.get(self.metadata['sample_fmt'], (0, 0))

        d['QuantizationBits'].value = bit_depth
        d['BlockAlign'].value = block_align

        d['Length'].value = self.length

        d['Compression'].value = AUID('04020202-0000-0000-060e-2b3404010101')
        return d

    def pixel_sizes(self):
        if not self.is_picture:
            return None

        pix_fmt = self.metadata['pix_fmt']
        h_samp = 2
        v_samp = 2

        depth = 8
        if pix_fmt.count('420'):
            h_samp = 2
            v_samp = 2
        elif pix_fmt.count('422'):
            h_samp = 2
            v_samp = 1
        elif pix_fmt.count('444'):
            h_samp = 1
            v_samp = 1

        for i in [8, 10, 12, 16]:
            if pix_fmt.count("p%d" % i):
                depth = i
                break

        return (depth, h_samp, v_samp)

    def get_avc_compression(self):
        if not self.is_picture:
            return None

        profile = self.metadata.get('profile', None)
        key = 'CompressedPicture'
        if profile == "Baseline":
            key = 'AVCBaselineUnconstrained'
        elif profile == "Constrained Baseline":
            key = 'AVCConstrainedBaselineUnconstrained'
        elif profile == "Main":
            key = 'AVCMainUnconstrained'
        elif profile == "Extended":
            key = 'AVCExtendedUnconstrained'
        elif profile == "High":
            key = 'AVCHighUnconstrained'
        elif profile == "High 10":
            key = 'AVCHigh10Unconstrained'
        elif profile == "High 10 Intra":
            key = 'AVCHigh10IntraUnconstrained'
        elif profile == "High 4:2:2":
            key = 'AVCHigh422Unconstrained'
        elif profile == "High 4:2:2 Intra":
            key = 'AVCHigh422IntraUnconstrained'
        elif profile == "High 4:4:4":
            # key = 'AVCHigh444IntraUnconstrained'
            key = 'CompressedPicture'
        elif profile == "High 4:4:4 Predictive":
            # key = 'AVCHigh444PredictiveUnconstrained'
            key = 'CompressedPicture'
        elif profile == "High 4:4:4 Intra":
            # key = 'AVCHigh444IntraUnconstrained'
            key = 'CompressedPicture'
        elif profile == 'CAVLC 4:4:4':
            # key = 'AVCCAVLC444IntraUnconstrained'
            key = 'CompressedPicture'

        return video.compression_ids[key]

    def get_compression(self):
        if not self.is_picture:
            return None

        codec_name = self.metadata.get('codec_name', None)
        if codec_name == 'mjpeg':
            return video.compression_ids['mjpeg']
        if codec_name == 'h264':
            return self.get_avc_compression()

        return video.compression_ids['CompressedPicture']

    def create_video_descriptor(self, f):
        if not self.is_picture:
            return None

        d = f.create.CDCIDescriptor()

        depth, h_samp, v_samp = self.pixel_sizes()

        width = self.metadata['width']
        height = self.metadata['height']

        aspect_ratio = "%d/%d" % (width, height)

        d['ComponentWidth'].value = depth
        d['HorizontalSubsampling'].value = h_samp
        d['VerticalSubsampling'].value = v_samp
        d['FrameLayout'].value = 'FullFrame'

        d['VideoLineMap'].value = [0, 0]
        # d['VideoLineMap'].value = [42, 0]
        d['ImageAspectRatio'].value = aspect_ratio

        d['StoredWidth'].value = width
        d['StoredHeight'].value = height
        d['SampleRate'].value = self.metadata['avg_frame_rate']
        compression = self.get_compression()

        d['Compression'].value = compression

        # d['ResolutionID'].value =  2900
        d['Length'].value = int(self.length)

        return d


def create_media_link(f, path, metadata):
    """
    Create an essence linked to external media and all obligatory mobs and data structures required by
    the edit spec.

    The returned :class:`aaf.mobs.MasterMob` will have one slot for each video stream and each audio channel
    in the file at `path`.

    Example: The linked file is a Quicktime movie with picture and a stereo audio track. This function will create a
    SourceMob with three slots, one picture slot, and two sound slots, for audio channels one and two respectively.
    The function will also create a derivation SourceMob, linked to these slots.

    :param f: The :class:`aaf.File` to add this link to
    :param path: A path recognizable to `os.path`
    :param metadata: Pre-fetched media description (in the form of a dictionary)
        from "ffprobe -show_format -show_streams"
    :return: A `aaf.mobs.MasterMob` linked to the file at link.
    """

    def tape_mob_for_format(name, format_info):
        tape_mob = f.create.SourceMob()
        tape_mob.name = name
        picture = format_info.first_picture_stream
        if picture is not None:
            pix_slot = tape_mob.create_picture_slot(edit_rate=picture.edit_rate)
            pix_slot.segment.length = picture.length
            tape_source_clip = f.create.SourceClip(media_kind='picture')
            tape_source_clip.length = picture.length
            pix_slot.segment.components.append(tape_source_clip)

        sound = format_info.first_sound_stream
        if sound is not None:
            for channel in range(sound.physical_track_count):
                sound_slot = tape_mob.create_sound_slot(edit_rate=sound.edit_rate)
                sound_slot.segment.length = sound.length
                tape_source_clip = f.create.SourceClip(media_kind='sound')
                tape_source_clip.length = sound.length
                sound_slot.segment.components.append(tape_source_clip)
                # not setting PhysicalTrackNumber here because we didn't before

        f.content.mobs.append(tape_mob)
        tape_mob.descriptor = f.create.ImportDescriptor()

        return tape_mob

    def append_source_to_mob_as_new_slots(from_mob, to_mob):
        sound_physical_track = 1
        for from_slot in from_mob.slots:
            slot_kind = from_slot.media_kind
            if slot_kind in ("Picture", "Sound"):
                from_clip = from_mob.create_source_clip(slot_id=from_slot.slot_id, media_kind=slot_kind)
                to_slot = to_mob.create_empty_sequence_slot(edit_rate=from_slot.edit_rate,
                                                            media_kind=from_clip.media_kind)
                to_slot.segment.components.append(from_clip)

                if slot_kind == 'Sound':
                    to_slot['PhysicalTrackNumber'].value = sound_physical_track
                    sound_physical_track += 1

    def source_mob_from_tape_mob(name, tape_mob):
        source_mob = f.create.SourceMob()
        source_mob.name = name

        append_source_to_mob_as_new_slots(from_mob=tape_mob, to_mob=source_mob)
        f.content.mobs.append(source_mob)

        return source_mob

    def master_mob_from_source_mob(name, source_mob):
        master_mob = f.create.MasterMob()
        master_mob.name = name

        append_source_to_mob_as_new_slots(from_mob=source_mob, to_mob=master_mob)

        f.content.mobs.append(master_mob)

        return master_mob

    def create_mobs(name, format_info):
        tmob = tape_mob_for_format(name + ' <TAPE MOB>', format_info)
        smob = source_mob_from_tape_mob(name + ' <SOURCE MOB>', tmob)
        mmob = master_mob_from_source_mob(name, smob)

        if format_info.first_picture_stream is not None:
            # MC Quicktime plugin will error if this is not set
            smob.comments['Video'] = format_info.first_picture_stream.codec_name

        return mmob, smob, tmob

    basename = os.path.basename(path)
    name, ext = os.path.splitext(basename)

    if ext == '.mxf' or ext == '.MXF':
        m = mxf.MXFFile(path)
        m.ama = True
        m.dump()
        return m.link(f)

    format_info = FormatInfo(metadata)
    source_descriptor = format_info.create_descriptor(f, path)

    if source_descriptor is None:
        return None

    master_mob, source_mob, tape_mob = create_mobs(name, format_info)
    source_mob.descriptor = source_descriptor
    return master_mob, source_mob, tape_mob



from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import itertools
import os
import io
import subprocess
import unittest
import hashlib
import common
import aaf2
from aaf2 import video, audio

class ImportTests(unittest.TestCase):

    def test_dnx_iter(self):
        profile_name = 'dnx_1080p_36_23.97'
        sample = common.generate_dnxhd(profile_name, "dnx_iter01.dnxhd", 10)
        with io.open(sample, 'rb', buffering=io.DEFAULT_BUFFER_SIZE) as f:
            for i, packet in enumerate(video.iter_dnx_stream(f)):
                cid, width, height, bitdepth, interlaced = video.read_dnx_frame_header(packet)
                assert not interlaced
                assert bitdepth == 8

                assert (width, height) == (1920, 1080)

            assert i+1 == 10

        profile_name = 'dnx_1080i_120_25'
        sample = common.generate_dnxhd(profile_name, "dnx_iter02.dnxhd", 10)
        with io.open(sample, 'rb', buffering=io.DEFAULT_BUFFER_SIZE) as f:
            for i, packet in enumerate(video.iter_dnx_stream(f)):
                cid, width, height, bitdepth, interlaced = video.read_dnx_frame_header(packet)
                assert interlaced
                assert bitdepth == 8
                assert (width, height) == (1920, 540)

            assert i+1 == 10

        frame_rate = '23.97'
        profile_name = 'dnxhr_lb'
        uhd2160 = (3840, 2160)
        sample = common.generate_dnxhd(profile_name, "dnx_iter03.dnxhd", 10, size=uhd2160, frame_rate=frame_rate)
        with io.open(sample, 'rb', buffering=io.DEFAULT_BUFFER_SIZE) as f:
            for i, packet in enumerate(video.iter_dnx_stream(f)):
                cid, width, height, bitdepth, interlaced = video.read_dnx_frame_header(packet)
                assert not interlaced
                assert (width, height) == uhd2160
                assert bitdepth == 8

            assert i+1 == 10

    def test_dnxhd(self):
        frames = 3
        for profile_name, offline in itertools.product(
                ['dnx_1080p_36_23.97', 'dnx_720p_90x_25', 'dnx_1080i_120_25', 'dnx_1080p_175x_23.97'],
                [True, False]
        ):
            new_file = os.path.join(common.sandbox(), '%s_import_essence.aaf' % profile_name)
            sample = common.generate_dnxhd(profile_name, '%s-import.dnxhd' % profile_name, frames=frames)

            with aaf2.open(new_file, 'w') as f:
                profile = video.dnx_profiles.get(profile_name)

                mob = f.create.MasterMob(profile_name)
                f.content.mobs.append(mob)
                mob.import_dnxhd_essence(sample, profile['frame_rate'], offline=offline)

            with aaf2.open(new_file, 'r') as f:
                mob = next(f.content.sourcemobs())
                if offline:
                    assert mob.essence is None
                else:
                    stream = mob.essence.open('r')
                    dump_path = os.path.join(common.sample_dir(),'%s-import-dump.dnxhd' % profile_name)
                    with io.open(dump_path, 'wb', buffering=io.DEFAULT_BUFFER_SIZE) as out:
                        out.write(stream.read())
                    assert common.compare_files(dump_path, sample)
                assert mob.slots[0].segment.length == frames
                assert mob.descriptor.length == frames

    def test_dnxhr(self):

        frame_rate = '23.97'
        uhd2160 = (960, 540)

        frames = 3

        timecode_fps = 30
        start_time = int(timecode_fps * 60 * 60) # 1 hour

        for profile_name in ['dnxhr_lb', 'dnxhr_sq', 'dnxhr_hq']:
            new_file = os.path.join(common.sandbox(), '%s_import_essence.aaf' % profile_name)
            with aaf2.open(new_file, 'w') as f:


                profile = video.dnx_profiles.get(profile_name)
                sample = common.generate_dnxhd(profile_name, "%s-import.dnxhd" % profile_name, frames=frames, size=uhd2160, frame_rate=frame_rate)

                # create a tape
                tape_mob = f.create.SourceMob()
                tape_mob.create_tape_slots(profile_name, frame_rate, timecode_fps)
                f.content.mobs.append(tape_mob)
                tape = tape_mob.create_source_clip(1, start=start_time)

                mob = f.create.MasterMob(profile_name)
                f.content.mobs.append(mob)
                timecode = f.create.Timecode(timecode_fps)
                mob.import_dnxhd_essence(sample, frame_rate, tape=tape)

            with aaf2.open(new_file, 'r') as f:
                source_mobs = []
                tape_mobs = []
                for mob in f.content.sourcemobs():
                    if isinstance(mob.descriptor, aaf2.essence.TapeDescriptor):
                        tape_mobs.append(mob)
                    else:
                        source_mobs.append(mob)

                mob = source_mobs[0]
                stream = mob.essence.open('r')
                dump_path = os.path.join(common.sample_dir(),'%s-import-dump.dnxhd' % profile_name)
                with io.open(dump_path, 'wb', buffering=io.DEFAULT_BUFFER_SIZE) as out:
                    out.write(stream.read())

                assert common.compare_files(dump_path, sample)
                assert mob.slots[0].segment.length == frames
                assert mob.slots[0].segment.start  == start_time
                assert mob.descriptor.length == frames


    def test_wav(self):
        # name, sample_rate = 48000, duration = 2, sample_fmt='s16le', format='wav'):
        # profile_name = 'pcm_48000_s24le'
        frame_rate = 30
        timecode_fps = 30
        start_time = int(timecode_fps * 60 * 60) * 2

        duration = 1.5
        for profile_name, offline in itertools.product(sorted(audio.pcm_profiles), [True, False]):
            sample_format = audio.pcm_profiles[profile_name]['sample_format']
            sample_rate = audio.pcm_profiles[profile_name]['sample_rate']

            sample = common.generate_pcm_audio_mono(profile_name, sample_format=sample_format, sample_rate=sample_rate, duration=duration)
            new_file = os.path.join(common.sandbox(), '%s_import_essence.aaf' % profile_name)
            with aaf2.open(new_file, 'w') as f:

                tape_mob = f.create.SourceMob()
                f.content.mobs.append(tape_mob)
                tape_mob.create_tape_slots(profile_name, frame_rate, timecode_fps, media_kind='sound')

                tape = tape_mob.create_source_clip(1, start_time)

                mob = f.create.MasterMob(profile_name)
                f.content.mobs.append(mob)
                mob.import_audio_essence(sample, frame_rate, tape, offline=offline)

            with aaf2.open(new_file, 'r') as f:
                source_mobs = []
                tape_mobs = []
                for mob in f.content.sourcemobs():
                    if isinstance(mob.descriptor, aaf2.essence.TapeDescriptor):
                        tape_mobs.append(mob)
                    else:
                        source_mobs.append(mob)
                mob = source_mobs[0]
                tape_mob = tape_mobs[0]
                if offline:
                    assert mob.essence is None
                else:
                    stream = mob.essence.open('r')
                    dump_path = os.path.join(common.sample_dir(),'%s-import-dump.wav' % profile_name)
                    mob.export_audio(dump_path)
                    assert common.compare_files(dump_path, sample)
                audio_samples = mob.descriptor['Length'].value
                assert audio_samples == sample_rate * duration
                edit_length = duration * frame_rate
                assert mob.slots[0].segment.length == edit_length
                assert mob.slots[0].segment.start == start_time
                assert mob.slots[0].segment.mob_id == tape_mob.mob_id

    def test_multi(self):
        frame_rate = 30.0
        frames = 4
        audio_duration = frames/frame_rate
        uhd2160 = (960, 540)
        new_file = os.path.join(common.sandbox(), 'multi_import_essence.aaf')
        audio_profile_name = 'pcm_48000_s24le'

        sample_format = audio.pcm_profiles[audio_profile_name]['sample_format']
        sample_rate = audio.pcm_profiles[audio_profile_name]['sample_rate']
        audio_sample = common.generate_pcm_audio_mono(audio_profile_name, sample_format=sample_format, sample_rate=sample_rate, duration=audio_duration)
        samples = {}
        mastermobs = {}

        timecode_fps = 30
        start_time = int(timecode_fps * 60 * 60)
        tape_id = None
        with aaf2.open(new_file, 'w') as f:
            # create a tape
            tape_mob = f.create.SourceMob()
            tape_mob.create_tape_slots("mulit-import", frame_rate, timecode_fps)
            tape_id = tape_mob.mob_id
            f.content.mobs.append(tape_mob)

            for i, profile_name in enumerate(['dnxhr_lb', 'dnxhr_sq', 'dnxhr_hq'], 1):
                profile = video.dnx_profiles.get(profile_name)
                sample = common.generate_dnxhd(profile_name, "%s-mulit-import.dnxhd" % profile_name, frames=frames, size=uhd2160, frame_rate=frame_rate)

                mob = f.create.MasterMob(profile_name)
                f.content.mobs.append(mob)

                tape = tape_mob.create_source_clip(1, start=start_time)

                vs_slot = mob.import_dnxhd_essence(sample, frame_rate, tape=tape)
                as_slot = mob.import_audio_essence(audio_sample, frame_rate)

                vs_mob = vs_slot.segment.mob
                as_mob = as_slot.segment.mob

                mob.comments['profile'] = profile_name
                mob.comments['integer'] = 100
                mob.comments['integer'] = i
                mastermobs[mob.mob_id] = (profile_name, i)

                v_dump_path = os.path.join(common.sample_dir(),'%s-multi-import-dump.wav' % profile_name)
                a_dump_path = os.path.join(common.sample_dir(),'%s-multi-import-dump.dnxhd' % profile_name)

                samples[vs_mob.mob_id] = (sample, v_dump_path, start_time)
                samples[as_mob.mob_id] = (audio_sample, a_dump_path, start_time)

                start_time += 100

        with aaf2.open(new_file, 'r') as f:
            tape_mob = f.content.mobs.get(tape_id)
            for mob_id, (profile_name, v) in mastermobs.items():
                mob = f.content.mobs.get(mob_id)
                assert mob.comments['profile'] == profile_name
                assert mob.comments['integer'] == v

            for mob_id, (original_path, dump_path, start_time) in samples.items():
                mob = f.content.mobs.get(mob_id)

                if isinstance(mob.descriptor, aaf2.essence.PCMDescriptor):
                    mob.export_audio(dump_path)
                    assert common.compare_files(dump_path, original_path)
                    assert mob.slots[0].segment.length == frames

                if isinstance(mob.descriptor, aaf2.essence.CDCIDescriptor):
                    stream = mob.essence.open('r')
                    with io.open(dump_path, 'wb', buffering=io.DEFAULT_BUFFER_SIZE) as out:
                        out.write(stream.read())
                    assert common.compare_files(dump_path, original_path)
                    assert mob.slots[0].segment.length == frames
                    assert mob.slots[0].segment.start == start_time
                    assert mob.slots[0].segment.mob_id == tape_id




if __name__ == "__main__":
    import logging
    # logging.basicConfig(level=logging.DEBUG)
    unittest.main()
